/**************************************************************************

  $Id: diag.c,v 1.3 2000/11/03 19:02:55 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.


  Program for calculation of G tensors from RHF data resulting from
  an AM1, PM3, MNDO or ZINDO/S RHF MO calculation using AMPAC or MOPAC
  or Hyperchem/Martin Platos addon (for ZINDO/S). Result is the the G
  tensor in the axis system the LCAO coefficients were calculated in
  as well as the eigenvalues and normalized eigenvectors.

  This file contains routines for calculation of eigenvalues and
  eigenvectors of real, symmetric matrices as well a routine for
  the calculation of error propagation. Most of this routines are
  derived from the book 'Numerical Recipes in C' by W. H. Press,
  S. A. Teukolsky, W. T. Vetterling and B. P. Flannery,
  Cambridge University Press, 1992 (2nd. edition).

  ATTENTION : Throughout this routines the definition of a matrix
			  follows the method described in "Numerical Recipes
			  in C" (see also the text in M_ALLOC.C)

**************************************************************************/


#include "diag.h"


#define MAX_ITERATIONS	30			/* maximum number of iterations to be */
									/* done while diagonalizing			  */

/*------------------------------------------------------------------------*/
/*	diagonalize() calculates the eigenvalues and eigenvalues of a real,	  */
/*	symmetric matrix. On return the eigenvectors are normalized to 1. The */
/*	original matrix is overwriten by the eigenvectors !					  */
/*																		  */
/*	->	 1. Dimension of the matrix										  */
/*		 2. Matrix to be diagonalized									  */
/*		 3. Array for the diagonal elements (eigenvalues)				  */
/*	<-	1 : everything ok, 0 : some kind of error happend				  */
/*		On return the eigenvalues are stored in the array 'eigenvalues',  */
/*		the eigenvectors are stored in the matrix 'a' (as its columns).	  */
/*------------------------------------------------------------------------*/


int diagonalize(int n, DMATRIX a, double *eigenvalues)
{
	double *off_diag;


	/* Get memory for the off-diagonal elements of the tridiagonal matrix */

	if ((off_diag = (double *) malloc(n * sizeof(double))) == NULL)
		return 0;

	/* Tridiagonalize the matrix, calculate eigenvalues and eigenvectors  */

	tridiagonalize(n, a, eigenvalues, off_diag);
	if (!diag_tridiag(n, eigenvalues, off_diag, a))
		return 0;

	return 1;
}


/*------------------------------------------------------------------------*/
/*	tridiagonalize() reduces a real, symmetric matrix to tridiagonal	  */
/*	form, i. e. a form, where only the diagonal elements and their		  */
/*	direct neighbours are non-zero. In this form it can be used by the	  */
/*	routine diag_tridiag() for further diagonalisation and calculation	  */
/*	of eigenvectors. The algorithm used is the Householder method. For	  */
/*	further information see `Numerical Recipes in C', p. 469 - 475.		  */
/*																		  */
/*	->	 1. Dimension of matrix											  */
/*		 2. Matrix to be tridiagonalised								  */
/*		 3. Array of diagonal elements									  */
/*		 4. Array of non-zero off-diagonal elements						  */
/*	<-	On return the diagonal elements are stored in the array 'd', the  */
/*		non-zero off-diagonal elements in 'e'. The first element of this  */
/*		array is always zero, there are only (n - 1) elements on the off- */
/*		diagonal directly beside the diagonal, since the resulting matrix */
/*		is also symmetric and only one array is necessary for the off-	  */
/*		diagonal elements. In the matrix 'a', the Householder rotation	  */
/*		matrix is stored, to be used by the routine diag_triadiag().	  */
/*------------------------------------------------------------------------*/


void tridiagonalize(int n, DMATRIX a, double *d, double *e)
{
	int i, j, k, l;
	double scale, h, hh, g, f;


	for (i = n - 1; i > 0; i--)
	{
		l = i - 1;
		h = scale = 0.0;

		if (i > 1)
		{
			for (k = 0; k < i; k++)
				scale += fabs(a[i][k]);

			if (scale == 0.0)
				e[i] = a[i][l];
			else
			{
				for (k = 0; k < i; k++)
				{
					a[i][k] /= scale;
					h += a[i][k] * a[i][k];
				}

				f = a[i][l];
				g = ((f >= 0.0) ? -sqrt(h) : sqrt(h));
				e[i] = scale * g;
				h -= f * g;
				a[i][l] = f - g;
				f = 0.0;

				for (j = 0; j < i; j++)
				{
					a[j][i] = a[i][j] / h;
					g = 0.0;

					for (k = 0; k <= j; k++)
						g += a[j][k] * a[i][k];

					for (k = j + 1; k < i; k++)
						g += a[k][j] * a[i][k];

					e[j] = g / h;
					f += e[j] * a[i][j];
				}

				hh = f / (h + h);

				for (j = 0; j < i; j++)
				{
					f = a[i][j];
					e[j] = g = e[j] - hh * f;

					for (k = 0; k <= j; k++)
						a[j][k] -= f * e[k] + g * a[i][k];
				}
			}
		}
		else
			e[i] = a[i][l];

		d[i] = h;
	}

	d[0] = e[0] = 0.0;

	for (i = 0; i < n; i++)
	{
		if (d[i] != 0.0)
		{
			for (j = 0; j < i; j++)
			{
				g = 0.0;
				for (k = 0; k < i; k++)
					g += a[i][k] * a[k][j];
				for (k = 0; k < i; k++)
					a[k][j] -= g * a[k][i];
			}
		}

		d[i] = a[i][i];
		a[i][i] = 1.0;
		for (j = 0; j < i; j++)
			a[j][i] = a[i][j] = 0.0;
	}
}


/*------------------------------------------------------------------------*/
/*	diag_tridiag() calculates the eigenvalues and eigenvectors of an	  */
/*	real, symmetric, tridiagonal matrix. For further information see	  */
/*	`Numerical Recipes in C', p. 475 - 481.								  */
/*	The maximum number of iterations can be adjusted via NUM_ITERATIONS,  */
/*	defined at the start of this file !									  */
/*																		  */
/*	->	 1. Dimension of the matrix										  */
/*		 2. Array of the diagonal elements of the matrix				  */
/*		 3. Array of the off-diagonal elements of the matrix (the first	  */
/*			entry in this array is arbitrary)							  */
/*		 4. When the matrix was reduced to tridiagonal form by the rou-	  */
/*			tine tridiagonalize(), this argument has to be the pointer	  */
/*			to the resulting Householder matrix, but if he matrix was	  */
/*			tridiagonal from the beginning, it has to be a pointer to	  */
/*			identity matrix	 !											  */
/*	<-	1 : everything ok, 0 : matrix can't be diagonalized				  */
/*		On return, the eigenvalues are stored in the array 'd' and the	  */
/*		matrix 'z' consists of the eigenvectors as its columns. The		  */
/*		eigenvectors are automatically normalized to 1.					  */
/*------------------------------------------------------------------------*/


int diag_tridiag(int n, double *d, double *e, DMATRIX z)
{
	int i, k, m, l, iter;
	double s, r, p, g, f, dd, c, b;


	for (i = 1; i < n; i++)
		e[i - 1] = e[i];
	e[n - 1] = 0.0;

	for (l = 0; l < n; l++)
	{
		iter = 0;
		do
		{
			for (m = l; m < n - 1; m++)
			{
				dd = fabs(d[m]) + fabs(d[m + 1]);
				if (fabs(e[m]) + dd == dd)
					break;
			}

			if (m != l)
			{
				if (iter++ == MAX_ITERATIONS)
					return 0;
				g = 0.5 * (d[l + 1] - d[l]) / e[l];
				r = pythag(g, 1.0);
				g = d[m] - d[l] + e[l] /
							  (g + ((g >= 0.0) ? fabs(r) : -fabs(r)));
				s = c = 1.0;
				p = 0.0;

				for (i = m - 1; i >= l; i--)
				{
					f = s * e[i];
					b = c * e[i];
					e[i + 1] = r = pythag(f, g);

					if (r == 0.0)
					{
						d[i + 1] -= p;
						e[m] = 0.0;
						break;
					}

					s = f / r;
					c = g / r;
					g = d[i + 1] - p;
					r = (d[i] - g) * s + 2.0 * c * b;
					p = s * r;
					d[i + 1]= g + p;
					g = c * r - b;

					for ( k = 0; k < n; k++)
					{
						f = z[k][i + 1];
						z[k][i + 1] = s * z[k][i] + c * f;
						z[k][i] = c * z[k][i] - s * f;
					}
				}

				if ((r == 0.0) && (i >= l))
					continue;

				d[l] -= p;
				e[l] = g;
				e[m] = 0.0;
			}
		} while (m != l);
	}

	return 1;
}


/*------------------------------------------------------------------------*/
/*	diag_errors() calculates the errors of the eigenvalues and eigen-	  */
/*	vector components, when the elements of the matrix, which has been	  */
/*	diagonalized by the routine diagonalize() have experimental errors.	  */
/*	The method was developed by W. H. Nelson, J. Mag. Res. 38, 71 (1980). */
/*																		  */
/*	->	 1. Dimension of matrix											  */
/*		 2. Matrix of errors											  */
/*		 3. Array of eigenvalues										  */
/*		 4. Array of errors of eigenvalues								  */
/*		 5. Matrix of eigenvectors as delivered by the diagonalisation	  */
/*			routine														  */
/*	<-	1 : everything ok, 0 : not enough memory						  */
/*		On return the errors of the eigenvalues are stored in the array	  */
/*		'e', while the errors of the eigenvectors are stored in the		  */
/*		matrix 'del'.													  */
/*------------------------------------------------------------------------*/


int diag_errors(int n, DMATRIX del, double *d, double *e, DMATRIX q)
{
	int i, j, k, l, m;
	DMATRIX x;
	double temp;


	/* Calculation of the error of the diagonal elements				  */

	for (i = 0; i < n; i++)
	{
		e[i] = 0.0;

		for (j = 0; j < n; j++)
		{
			for (k = j; k < n; k++)
			{
				if (k == j)
					e[i] += q[j][i] * q[j][i] * q[j][i] * q[j][i]
													* del[j][j] * del[j][j];
				else
					e[i] += 4.0 * q[j][i] * q[j][i] * q[k][i] * q[k][i]
													* del[j][k] * del[j][k];
			}
		}

		e[i] = sqrt(e[i]);
	}

	/* Calculation of the errors of the eigenvectors					  */

	if ((x = dmatrix(n, n)) == NULL)	/* temporary matrix for errors	  */
		return 0;

	for (i = 0; i < n; i++)				/* loop over all eigenvectors	  */
	{
		for (j = 0; j < n; j++)			/* loop over all components		  */
		{
			x[i][j] = 0.0;

			for (k = 0; k < n; k++)
			{
				for (l = k; l < n; l++)
				{
					temp = 0.0;

					for (m = 0; m < n; m++)
					{
						if ((m == i) || (d[i] == d[m]))
							continue;

						if (k == l)
							temp += q[k][m] * q[k][i] * q[j][m] /
															  (d[i] - d[m]);
						else
							temp += (q[k][m] * q[l][i] + q[l][m] * q[k][i])
												  * q[j][m] / (d[i] - d[m]);
					}

					x[i][j] += temp * temp * del[k][l] * del[k][l];
				}
			}
		}
	}

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			del[j][i] = sqrt(x[i][j]); 

	free(x);
	return 1;
}


/*--------------------------------------------------------------------*/
/*	sort_eigen_desc() sorts the eigenvalues in the array 'd' in des-  */
/*  cending order and rearrages the columns of the matrix 'a' (i. e.  */
/*  the eigenvectors) correspondingly.								  */
/*																	  */
/*	->	 1. Dimension of matrix = number of eigenvalues				  */
/*		 2. Array of eigenvalues									  */
/*		 3. Matrix of eigenvectors (as columns of the matrix)		  */
/*--------------------------------------------------------------------*/


void sort_eigen_desc(int n, double *d, DMATRIX a)
{
	int i, j, k;
	double temp;


	for (i = 0; i < n; i++)
	{
		k = i;
		temp = d[k];

		for (j = i + 1; j < n; j++)
		{
			if (d[j] > temp)
			{
				k = j;
				temp = d[k];
			}
		}

		if (k != i)
		{
			d[k] = d[i];
			d[i] = temp;

			for (j =0; j < n; j++)
			{
				temp = a[j][i];
				a[j][i] = a[j][k];
				a[j][k] = temp;
			}
		}
	}
}


/*------------------------------------------------------------------------*/
/*	sort_eigen_asc() sorts the eigenvalues in the array 'd' in ascending  */
/*	order and rearrages the columns of the matrix 'a' (i. e. the eigen-	  */
/*	vectors) correspondingly.											  */
/*																		  */
/*	->	 1. Dimension of matrix = number of eigenvalues					  */
/*		 2. Array of eigenvalues										  */
/*		 3. Matrix of eigenvectors (as columns of the matrix)			  */
/*------------------------------------------------------------------------*/


void sort_eigen_asc(int n, double *d, DMATRIX a)
{
	int i, j, k;
	double temp;


	for (i = 0; i < n; i++)
	{
		k = i;
		temp = d[k];

		for (j = i + 1; j < n; j++)
		{
			if (d[j] < temp)
			{
				k = j;
				temp = d[k];
			}
		}

		if (k != i)
		{
			d[k] = d[i];
			d[i] = temp;

			for (j =0; j < n; j++)
			{
				temp = a[j][i];
				a[j][i] = a[j][k];
				a[j][k] = temp;
			}
		}
	}
}


/*------------------------------------------------------------------------*/
/*	pythag() computes sqrt(a * a + b * b) without destructive over- or	  */
/*	underflow. For further information see `Numerical Recipes is C',	  */
/*	p. 70.																  */
/*																		  */
/*	->	Two double values												  */
/*	<-	sqrt(a * a + b * b) as double value								  */
/*------------------------------------------------------------------------*/


double pythag( double a, double b )
{
	double abs_a, abs_b;


	abs_a = fabs( a );
	abs_b = fabs( b );
	if ( abs_a > abs_b )
		return abs_a * sqrt( 1.0 + abs_b * abs_b / ( abs_a * abs_a ) );
	else
		return abs_b == 0.0 ?
			0.0 : abs_b * sqrt( 1.0 + abs_a * abs_a / ( abs_b * abs_b ) );
}
