/*********************************************************************

  $Id: e_corr.c,v 1.5 2000/11/03 18:46:19 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  This file contains all routines used in the calculation of energy
  corrections needed for the calculation of G tensors.

*********************************************************************/


#include "g_rhf.h"


/* Declaration of routiones local to this file */

static int get_memory_for_LCAOs( DMATRIX *clK, DMATRIX *clL, int states );
static void rotate_in_local_system( ATOM *aK, ATOM *aL, DMATRIX clK,
									DMATRIX clL, double R, int states );
static void make_JK_1( double *JI, double *KI, DMATRIX c,
					   double tc_int[ 4 ][ 4 ][ 4 ][ 4 ],
					   int slp, int elp, int i, int j );
static void make_JK_2( double *JI, double *KI, DMATRIX cK, DMATRIX cL,
					   double tc_int[ 4 ][ 4 ][ 4 ][ 4 ], int slp, int elp,
					   int i, int j );


/* Definitions needed for this file */

#define q_q       0                 /* different types of multipole- */
#define q_uz      1                 /* multipole interactions */
#define uz_q      2
#define q_Qpp     3
#define Qpp_q     4
#define q_Qzz     5
#define Qzz_q     6
#define up_up     7
#define uz_uz     8
#define up_Qpz    9
#define Qpz_up   10
#define uz_Qpp   11
#define Qpp_uz   12
#define uz_Qzz   13
#define Qzz_uz   14
#define Qpp_Qpp  15
#define Qxx_Qyy  16
#define Qpp_Qzz  17
#define Qzz_Qpp  18
#define Qzz_Qzz  19
#define Qpz_Qpz  20
#define Qxy_Qxy  21


/* Here follows a list of all combinations of atomic orbitals leading to
   non-vanishing two-center integrals. They are define globally since they
   are needed in several routines in this file.
   There are 256 possible combinations of non-vanishing two-center
   integrals but only 72 of them are non-zero. So instead of running
   through all possible integrals we calculate only the non-vanishing
   ones and index them with the values from the following array, thus
   reducing the computation time.
   - The integral for the first combination of atomic orbitals, tcidx[0],
     has to be calculated every time, i.e. for every combination of atoms
   - the integrals indexed by  tcidx[1] to tcidx[5] have to be calculated
     if the first atom has a quantum number of n > 1
   - the integrals for tcidx[6] to tcidx[10] have to be calculated if
     the second atom has a quantum number of n > 1
   - the integrals for tcidx[11] to tcidx[71] have to be calculated only
     when both atoms have a quantum nuber of n > 1. */

int tcidx[ 72 ][ 4 ] = {                       /* always                */
                         {  S,  S,  S,  S },   /*   0.  (S S, S S)      */
                                               /* first atom has n > 1  */
                         { PX, PX,  S,  S },   /*  1.  (Pp Pp, S S)     */
                         { PY, PY,  S,  S },   /*  2.                   */
                         { PZ, PZ,  S,  S },   /*  3.  (Pz Pz, S S)     */
                         {  S, PZ,  S,  S },   /*  4.  (S Pz, S S)      */
                         { PZ,  S,  S,  S },   /*  5.                   */
                                               /* second atom has n > 1 */
                         {  S,  S, PX, PX },   /*  6.  (S S, Pp Pp)     */
                         {  S,  S, PY, PY },   /*  7.                   */
                         {  S,  S, PZ, PZ },   /*  8.  (S S, Pz Pz)     */
                         {  S,  S,  S, PZ },   /*  9.  (S S, S Pz)      */
                         {  S,  S, PZ,  S },   /* 10.                   */
                                               /* both atoms have n > 1 */
                         {  S, PX,  S, PX },   /* 11.  (S Pp, S Pp)     */
                         {  S, PX, PX,  S },   /* 12.                   */
                         { PX,  S,  S, PX },   /* 13.                   */
                         { PX,  S, PX,  S },   /* 14.                   */
                         {  S, PY,  S, PY },   /* 15.                   */
                         {  S, PY, PY,  S },   /* 16.                   */
                         { PY,  S,  S, PY },   /* 17.                   */
                         { PY,  S, PY,  S },   /* 18.                   */
                         {  S, PZ,  S, PZ },   /* 19.  (S Pz, S Pz)     */
                         {  S, PZ, PZ,  S },   /* 20.                   */
                         { PZ,  S,  S, PZ },   /* 21.                   */
                         { PZ,  S, PZ,  S },   /* 22.                   */
                         {  S, PX, PX, PZ },   /* 23.  (S Pp, Pp Pz)    */
                         {  S, PX, PZ, PX },   /* 24.                   */
                         { PX,  S, PX, PZ },   /* 25.                   */
                         { PX,  S, PZ, PX },   /* 26.                   */
                         {  S, PY, PY, PZ },   /* 27.                   */
                         {  S, PY, PZ, PY },   /* 28.                   */
                         { PY,  S, PY, PZ },   /* 29.                   */
                         { PY,  S, PZ, PY },   /* 30.                   */
                         { PX, PZ,  S, PX },   /* 31.  (Pp Pz, S Pp)    */
                         { PX, PZ, PX,  S },   /* 32.                   */
                         { PZ, PX,  S, PX },   /* 33.                   */
                         { PZ, PX, PX,  S },   /* 34.                   */
                         { PY, PZ,  S, PY },   /* 35.                   */
                         { PY, PZ, PY,  S },   /* 36.                   */
                         { PZ, PY,  S, PY },   /* 37.                   */
                         { PZ, PY, PY,  S },   /* 38.                   */
                         {  S, PZ, PX, PX },   /* 39.  (S PZ, Pp Pp)    */
                         { PZ,  S, PX, PX },   /* 40.                   */
                         {  S, PZ, PY, PY },   /* 41.                   */
                         { PZ,  S, PY, PY },   /* 42.                   */
                         { PX, PX,  S, PZ },   /* 43.  (Pp pP, S Pz)    */
                         { PX, PX, PZ,  S },   /* 44.                   */
                         { PY, PY,  S, PZ },   /* 45.                   */
                         { PY, PY, PZ,  S },   /* 46.                   */
                         {  S, PZ, PZ, PZ },   /* 47.  (S Pz, Pz Pz)    */
                         { PZ,  S, PZ, PZ },   /* 48.                   */
                         { PZ, PZ,  S, PZ },   /* 49.  (Pz Pz, S Pz)    */
                         { PZ, PZ, PZ,  S },   /* 50.                   */
                         { PX, PX, PX, PX },   /* 51.  (Pp Pp, Pp Pp)   */
                         { PY, PY, PY, PY },   /* 52.                   */
                         { PX, PX, PY, PY },   /* 53.  (Px Px, Py Py)   */
                         { PY, PY, PX, PX },   /* 54.                   */
                         { PX, PX, PZ, PZ },   /* 55.  (Pp Pp, Pz Pz)   */
                         { PY, PY, PZ, PZ },   /* 56.                   */
                         { PZ, PZ, PX, PX },   /* 57.  (Pz Pz, Pp Pp)   */
                         { PZ, PZ, PY, PY },   /* 58.                   */
                         { PZ, PZ, PZ, PZ },   /* 59.  (Pz Pz, Pz Pz)   */
                         { PX, PZ, PX, PZ },   /* 60.  (Pp Pz, Pp Pz)   */
                         { PX, PZ, PZ, PX },   /* 61.                   */
                         { PZ, PX, PX, PZ },   /* 62.                   */
                         { PZ, PX, PZ, PX },   /* 63.                   */
                         { PY, PZ, PY, PZ },   /* 64.                   */
                         { PY, PZ, PZ, PY },   /* 65.                   */
                         { PZ, PY, PY, PZ },   /* 66.                   */
                         { PZ, PY, PZ, PY },   /* 67.                   */
                         { PX, PY, PX, PY },   /* 68.  (Px Py, Px Py)   */
                         { PX, PY, PY, PX },   /* 69.                   */
                         { PY, PX, PX, PY },   /* 70.                   */
                         { PY, PX, PY, PX } }; /* 71.                   */



/************************************************************************/
/* Does everything necessary for the energy corrections, i.e. allocates */
/* memory, calculates the corrections and logs them.                    */
/************************************************************************/

int energy_corrections( double **JI, double **KI, ATOM *atom, double *energy,
						NUMS *n, OPTION *options )
{
	if ( ! options->dec )
		return 1;

	if ( ! get_memory_for_corrections( JI, KI, n ) )
		return 0;

	if ( ! calc_energy_corr( *JI, *KI, atom, n ) )
		return 0;

	if ( options->le )
		log_energy_corrections( options->outfile, energy, *JI, *KI,
								n, options );

	return 1;
}


/**************************************************************************/
/* Allocates memory used in calculation of Coulomb and exchange integrals */
/* Input parameter : Pointer to arrays of Coulomb and exchange integrals  */
/*                   Structure for number of atoms, states etc.           */
/* Return value : 1 = everything ok, 0 = failure                          */
/**************************************************************************/

int get_memory_for_corrections( double **JI, double **KI, NUMS *n )
{
    if ( ( *JI = ( double * ) calloc( 2 * n->states, sizeof( double ) ) )
		 == NULL )
        return 0;

	*KI = *JI + n->states;
    return 1;
}


/****************************************************************/
/* Deallocates memory used in calculation of energy corrections */
/* Input parameter : Pointer to array of Coulomb integrals      */
/****************************************************************/

void free_energy_corr_memory( double *JI )
{
	if ( JI != NULL )
		free( JI );
}


/**********************************************************************/
/* Calculates the Coulomb and exchange integals needed for the energy */
/* corrections to get transition energies from the obital energies.   */
/* Input parameter : arrays for Coulomb and exchange integrals        */
/*                   pointer to the array of structures for all atoms */
/*                   structure for number of atoms, states etc.       */
/* Return value : 1 = everything ok, 0 = failure                      */
/**********************************************************************/

int calc_energy_corr( double *JI, double *KI, ATOM *atom, NUMS *n )
{
    int i;                               /* counter over states */
    int K, L;                            /* counters over atoms */
    DMATRIX clK, clL;                    /* LCAO coefficients in the local */
                                         /* interatomic coordinate system */
    double R;                            /* interatomic distance */
    double m_int[ 22 ],                  /* multipole-multipole-integrals */
           tc_int[ 4 ][ 4 ][ 4 ][ 4 ];   /* and two-center-integrals */



    /* First allocate some memory for the LCAO coefficients in the local
       interatomic system */

    if ( ! get_memory_for_LCAOs( &clK, &clL, n->states ) )
    {
        free_energy_corr_memory( JI );
        return 0;
    }

    /* To calculate the corrections, we've got to sum over all pairs of
       atoms K and L (actually L is for symmetry reasons only running from
       K to the end (or possibly from 0 up to and including K), making
       things look a bit different in the routine make_JK_2) */

    for ( K = 0; K < n->atoms; K++ )
        for ( L = 0; L <= K; L++ )
        {
            /* We start with case that the atoms K and L are the same atom,
               then the calculation is rather simple and can be done in the
               normal coordinate system */

            if ( K == L )
            {
				/* Due to the bogosity of the approximations used the one-
				   center two-electron interactions are not the same as the
				   two-center two-electron interactions with the interatomic
				   distance set to 0. So we've got to calculate them seperately
				   from this other interactions... */

				make_oc_ints( tc_int, &atom[ K ] );

				/* Using this data the one-center Coulomb and exchange inter-
				   actions are computed */

                for ( i = 0; i < n->states; i++ )
				{
					make_JK_1( &JI[ i ], &KI[ i ], atom[ K ].c, tc_int, 0, 0,
							   i, n->SOMO );

					if ( atom[ K ].n == 1 )     /* ... already finished ? */
						continue;

					make_JK_1( &JI[ i ], &KI[ i ], atom[ K ].c, tc_int, 1, 71,
							   i, n->SOMO );
				}

                continue;         /* finished for the case K == L */
            }

            /* Otherwise we calulate the 22 two-center two-electron multi-
			   pole-multipole integrals for the actual combination of atoms
			   K and L, and from them the 72 non-vanishing two-center
			   integrals (but only as many as really are needed). For this
			   we also need the inter- atomic distance R. */

            R = sqrt(   sqr( atom[ L ].x - atom[ K ].x )
					  + sqr( atom[ L ].y - atom[ K ].y )
					  + sqr( atom[ L ].z - atom[ K ].z ) );

            make_integrals_1( m_int, atom, R, K, L );
            make_integrals_2( m_int, tc_int, atom[ K ].n, atom[ L ].n );

            /* We also have to transformate the LCAO coefficients into the
               local coordinates system */

            rotate_in_local_system( &atom[ K ], &atom[ L ], 
									clK, clL, R, n->states );

            /* And then we calculate the two-center contributions to the
               Coulomb and exchange terms */

			for ( i = 0; i < n->states; i++ )
			{
				make_JK_2( &JI[ i ], &KI[ i ], clK, clL, tc_int,
						   0, 0, i, n->SOMO );

				if ( atom[ K ].n > 1 )             /* atom K has n > 1 */
					make_JK_2( &JI[ i ], &KI[ i ], clK, clL, tc_int, 1, 5,
							   i, n->SOMO );

				if ( atom[ L ].n > 1 )             /* atom L has n > 1 */
					make_JK_2( &JI[ i ], &KI[ i ], clK, clL, tc_int, 6, 10,
							   i, n->SOMO );

				/* If one of the atoms has n = 1 we're finished, otherwise we
				   also got to calculate the rest of the multipole integrals */

				if ( atom[ K ].n == 1 || atom[ L ].n == 1 )
					continue;

				make_JK_2( &JI[ i ], &KI[ i ], clK, clL, tc_int, 11, 71, i,
						   n->SOMO );

			}         /* end of loop over all states */
        }           /* end of loop over all pairs of atoms K and L */

    /* Now we are finished and all left to be done is to deallocate memory not
       used anymore */

    free( clK );
    free( clL );

    return 1;
}


/*****************************************************************************/
/* Allocates memory for the LCAO coefficients in the local coordinate system */
/* Input parameter : pointer to the matrices of the LCAO coefficients        */
/*                   number of states                                        */
/* Return value : 1 = everything ok, 0 = failure                             */
/*****************************************************************************/

int get_memory_for_LCAOs( DMATRIX *clK, DMATRIX *clL, int states )
{
    if ( ( *clK = dmatrix( 4, states ) ) == NULL )
        return 0;

    if ( ( *clL = dmatrix( 4, states ) ) == NULL )
    {
        free( *clK );
        return 0;
    }

    return 1;
}


/****************************************************************/
/* Rotates the LCAO coefficients of all molecular orbitals into */
/* the approproate local coordinate system                      */
/* Input parameter : pointer to atom structure of first atom    */
/*                   pointer to atom structure of second atom   */
/*                   matrices for rotated alpha and beta LCAOs  */
/*                   of both atoms                              */
/*                   interatomic distance                       */
/*                   total number of alpha and beta orbitals    */
/****************************************************************/

static void rotate_in_local_system( ATOM *aK, ATOM *aL, DMATRIX clK,
									DMATRIX clL, double R, int states )
{
    DMATRIX cK, cL;                 /* temporary LCAO coefficient matrices */
    double sp, cp, st, ct;          /* sines and cosines of rotation angles */
    double temp;                    /* temporary variable */
    int i;                          /* counter for orbitals */


    cK = aK->c;                     /* using temporary variables to make */
    cL = aL->c;                     /* access just a bit faster          */


    /* Before two-center contributions can be calculated the atomic orbitals
       have to be rotated from the normal to a local interatomic coordinate
       system where the z-axis is pointing from atom K to atom L. This
       rotation can be seen as a rotation by an angle phi around the old
       z-axis, followed by a rotation around the new y'-axis by an angle
       theta. This is done by multiplying the "vector"

	                       / c[px][i] \
	                      |  c[py][i]  |
		                   \ c[pz][i] /

	   in the basis of the `unit vectors' |px>, |py>, |pz> by the matrix R

             / cos(phi) cos(theta)   sin(phi) cos(theta)   -sin(theta) \
        R = |     -sin(phi)                cos(phi)            0        |
             \ cos(phi) sin(theta)   sin(phi) sin(theta)    cos(theta) /

        thereby getting the LCAO coefficients in the `local basis'

          c[px'][i] * |px'> + c[py'][i] * |py'> + c[pz'][i] * |pz'>

        But first the sines und cosines of the angles phi and theta have
        to be calculate ... */

    ct = ( aL->z - aK->z ) / R;
    st = sqrt( 1.0 - sqr( ct ) );

    if ( fabs( ct ) > 0.9999999 )     /* both atoms already along the z-axis */
    {
        cp = 1.0;
        sp = 0.0;
    }
    else
    {
        temp = 1.0 / sqrt( sqr( aL->x - aK->x ) + sqr( aL->y - aK->y ) );

        cp = ( aL->x - aK->x ) * temp;
        sp = ( aL->y - aK->y ) * temp;
    }

    /* Now comes the rotation of the LCAO coefficients for all alpha
	   and beta states on the atoms K and L */

    for (i = 0; i < states; i++)
    {
        clK[  S ][ i ] =     cK[  S ][ i ];
        clK[ PX ][ i ] = (   cK[ PX ][ i ] * cp
                           + cK[ PY ][ i ] * sp ) * ct
                           - cK[ PZ ][ i ] * st;
        clK[ PY ][ i ] =   - cK[ PX ][ i ] * sp
                           + cK[ PY ][ i ] * cp;
        clK[ PZ ][ i ] = (   cK[ PX ][ i ] * cp
                           + cK[ PY ][ i ] * sp ) * st
                           + cK[ PZ ][ i ] * ct;

        clL[  S ][ i ] =     cL[  S ][ i ];
        clL[ PX ][ i ] = (   cL[ PX ][ i ] * cp
                           + cL[ PY ][ i ] * sp ) * ct
                           - cL[ PZ ][ i ] * st;
        clL[ PY ][ i ] =   - cL[ PX ][ i ] * sp
                           + cL[ PY ][ i ] * cp;
        clL[ PZ ][ i ] = (   cL[ PX ][ i ] * cp
                           + cL[ PY ][ i ] * sp ) * st
                           + cL[ PZ ][ i ] * ct;

   }  /* end of loop over all states */
}


/**********************************************************************/
/* Calculates the contribution to the Coulomb and exchange integrals  */
/* JI and JK from the one-center multipole interactions indexed by    */
/* slp to elp.                                                        */
/* Input parameter : pointer to actual JI and KI                      */
/*                   matrices of alpha and beta LCAOs of actual atom  */
/*                   array of multipole-multipole integrals           */
/*                   start end end value to be used in index array    */
/*                   numbers of states between the which Coulomb and  */
/*                   exchange integrals are to be calculated          */
/**********************************************************************/

static void make_JK_1( double *JI, double *KI, DMATRIX c,
					   double tc_int[ 4 ][ 4 ][ 4 ][ 4 ],
					   int slp, int elp, int i, int j )
{
    int k, m, n, l, s;
    double temp, act_tc_int;


    for ( k = slp; k <= elp; k++ )
    {
        m = tcidx[ k ][ 0 ];     /* mu     */
        n = tcidx[ k ][ 1 ];     /* nu     */
        l = tcidx[ k ][ 2 ];     /* lambda */
        s = tcidx[ k ][ 3 ];     /* sigma  */

        act_tc_int = tc_int[ m ][ n ][ l ][ s ];

        temp = act_tc_int * c[ m ][ i ] * c[ l ][ j ];
        *JI += temp * c[ n ][ i ] * c[ s ][ j ];
        *KI += temp * c[ n ][ j ] * c[ s ][ i ];
    }
}


/*********************************************************************/
/* Calculates the contribution to th Coulomb and exchange integrals  */
/* JI and KI from two-center multipole interactions indexed by slp   */
/* to elp.                                                           */
/* Input parameter : pointer to actual JI and KI                     */
/*                   matrices of alpha and beta LCAOs of both the    */
/*                   actual atoms                                    */
/*                   array of multipole-multipole integrals          */
/*                   start end end value to be used in index array   */
/*                   numbers of states between which the Coulomb and */
/*                   exchange integrals are to be calculated         */
/*********************************************************************/

static void make_JK_2( double *JI, double *KI, DMATRIX cK, DMATRIX cL,
					   double tc_int[ 4 ][ 4 ][ 4 ][ 4 ], int slp, int elp,
					   int i, int j )
{
    int k, m, n, l, s;
    double act_tc_int, temp1, temp2, temp3, temp4;


    for ( k = slp; k <= elp; k++ )
    {
        m = tcidx[ k ][ 0 ];     /* mu     */
        n = tcidx[ k ][ 1 ];     /* nu     */
        l = tcidx[ k ][ 2 ];     /* lambda */
        s = tcidx[ k ][ 3 ];     /* sigma  */

        act_tc_int = tc_int[m][n][l][s];

        temp1 = cK[m][i] * cL[ l ][ j ];
        temp2 = cK[m][j] * cL[ l ][ i ];
        temp3 = cK[n][i] * cL[ s ][ j ];
        temp4 = cK[n][j] * cL[ s ][ i ];
        *JI  += act_tc_int * ( temp1 * temp3 + temp2 * temp4 );
        *KI  += act_tc_int * ( temp1 * temp4 + temp2 * temp3 );
    }
}
