/*
  $Id: g_rhf.c,v 1.4 2001/08/19 20:57:49 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.


  Program for calculation of G tensors from RHF data resulting from
  an AM1, PM3, MNDO or ZINDO/S RHF MO calculation using AMPAC or MOPAC
  or Hyperchem/Martin Platos addon (for ZINDO/S). Result is the the G
  tensor in the axis system the LCAO coefficients were calculated in
  as well as the eigenvalues and normalized eigenvectors.

  Beside this file there are some more files needed:

 	1. g_rhf.h         Global definitions and declarations
 	2. g_util.c        Utility routines
 	3. m_alloc.c       Memory allocation routines
 	4. m_alloc.h       Declarations for m_alloc.c
 	5. e_corr.c        Routines for calculation of energy corrections
 	6. e_util.c        Utility routines for e_corr.c
 	7. log.c           Output routines
 	8. read.c          Routines for input of data file
 	9. paramag.c       Routines for calculation of paramagnetic part
   10. param_u.c       Utility routines for paramag.c
   11. relativ.c       Routines for calculation of relativistic part
   12. diamag.c        Routines for calculation of diamagnetic part
   13. diag.c          Matrix diagonalization routines
   14. diag.h          Declarations for diag.c
   15. matrix.c        Routines for calculation of determinants etc.
   16. matrix.h        Declarations for matrix.c
   17. param.pl        Parses input file for type of parameterization
   18. params.pl       Parses file with parameterization data
   19. numbers.pl      Parses input file for number of atoms, orbitals etc.
   20. coordina.pl     Parses input file for atomic coordinates
   21. energies.pl     Parses input file for orbital energies
   22. lcaos.pl        Parses input file for LCAO coefficients
   23. spin_orbit.pl   Parses file with spin-orbit coupling constants
   24. g_lpsort        Script for sorting log file
   25. AM1             Parameterization data for AM1
   25. PM3             Parameterization data for PM3
   26. MNDO            Parameterization data for MNDO
   27. ZINDOS          Parameterization data for ZINDO/S
   28. spin_orbit      Spin-orbit coupling constants

************************************************************************/


#include "g_rhf.h"            /* definitions and declarations */



/* Declaration of routines local to this file */


static int diagonalize_tensor( DMATRIX g, DMATRIX g_diag, double *g_eigen,
							   DMATRIX g1, DMATRIX g2, DMATRIX g3, DMATRIX dg,
							   DMATRIX dgt );
static void free_memory( DMATRIX g, DMATRIX g_diag, DMATRIX g1, DMATRIX g2,
						 DMATRIX g3, DMATRIX dg, DMATRIX dgt, ATOM *atom,
						 double *alpha_energy, int atoms );



/*********************************************/
/*             Here we start ...             */
/*********************************************/


int main( int argc, char **argv )
{
	NUMS n;                     /* structure for number of atoms, states etc */
    int diag_ok;                /* flag, set when diagonalization was ok */
	OPTION options;             /* structure of option flags */
    ATOM *atom;                 /* array of structures for each atom */
    double *energy;             /* array of orbital energies of MO-states */
    PARAMS params[ MAX_PARAMS ];/* structures for atomic parameters */
    DMATRIX g1, g2, g3,         /* all different components of the G tensor*/
            g, g_diag,
	        dg, dgt;
    double g_eigen[ 3 ];        /* eigenvalues of G tensor */
	clock_t start_time;         /* time of start of program */


    /**********************************************/
    /* First part: memory for all needed data is */
	/* allocated and data files are read in      */
    /*********************************************/

	/* First thing, get the time at the start of the program */

	start_time = clock( );

	/* Open data file, get name of parameter file and read it in */

	if ( ! open_data_file( argc, argv, params, &options ) )
		return 0;

	/* Read in all needed data from input data file */

	if ( ! get_all_data( &n, &atom, &energy, params, &options ) )
		return 0;

    /* Finally allocate memory for the different components of the
       G tensor and initialize them */

	if ( ! set_up_g_matrices( &g, &g_diag, &g1, &g2, &g3, &dg, &dgt, energy,
							  atom, n.atoms ) )
		return 0;


    /******************************************************/
    /* Second part of program: the G tensor is calculated */
    /******************************************************/

    /* Calculate paramagnetic, relativistic and diamagnetic parts */

	if ( ! calculate_paramagnetic_part( g3, atom, energy, &n, &options ) )
	{
		printf( "Not enough memory for calculation of energy "
				"corrections !\n" );
		free_memory( g, g_diag, g1, g2, g3, dg, dgt, atom, energy, n.atoms );
		return 0;
	}

	if ( ! options.Stone )
	{
		calculate_relativistic_part( g1, atom, &n );
		if ( options.Angstl )
			calculate_diamagnetic_part_Angstl( g2, atom, &n );
		else
			calculate_diamagnetic_part( g2, atom, &n );
	}


    /**********************************************************************/
    /* Third part: the final G tensor is diagonalized and written to file */
    /**********************************************************************/

	diag_ok = diagonalize_tensor( g, g_diag, g_eigen, g1, g2, g3, dg, dgt );
	write_results( diag_ok, g, g_diag, g_eigen, dg, g1, g2, g3, &options,
				   start_time );


    /***********************************************/
    /* Fourth and last part: memory is deallocated */
    /***********************************************/

	free_memory( g, g_diag, g1, g2, g3, dg, dgt, atom, energy, n.atoms );
    return 0;
}



/*********************************************************/
/* Calculates the symmetric G tensor and diagonalizes it */
/* Take care: On return the eigenvalues are stored in    */
/*            `g_eigen' while the eigenvectors are       */
/*            written to `g_diag' !                      */
/*********************************************************/

static int diagonalize_tensor( DMATRIX g, DMATRIX g_diag, double *g_eigen,
							   DMATRIX g1, DMATRIX g2, DMATRIX g3,
							   DMATRIX dg, DMATRIX dgt )
{
	int l, m, k;


    /* First calculate the complete G tensor */

    for ( l = X; l <= Z; l++ )
        for ( m = X; m <= Z; m++ )
		{
			g1[ l ][ m ] *= g_e * sqr( alpha );
			g2[ l ][ m ] *= g_e * sqr( alpha );
			g3[ l ][ m ] *= g_e * sqr( alpha );
			g[ l ][ m ] += g1[ l ][ m ] + g2[ l ][ m ] + g3[ l ][ m ];
		}

    /* Calculate G times its transposed */

    for ( l = X; l <= Z; l++ )
		for ( m = X; m <= Z; m++ )
		{
			g_diag[ l ][ m ] = 0.0;

			for ( k = X; k <= Z; k++ )
				g_diag[ l ][ m ] += g[ l ][ k ] * g[ m ][ k ];
		}

    /* Diagonize the square of the G tensor */

    if ( ! diagonalize( 3, g_diag, g_eigen ) )
	{
		printf( "Diagonalization of square of G tensor failed!\n" );
		return 0;
	}

	/* Calculate the eigenvalues of the G tensor itself */

	for ( k = X; k <= Z; k++ )
	{
		if ( g_eigen[ k ] < 0.0 )
		{
			printf( "%1d. eigenvalue of square of G tensor is negative!\n",
					k + 1 );
			return 0;
		}

		g_eigen[ k ] = sqrt( g_eigen[ k ] );
	}

	/* Finally rotate the G tensor back into the molecular axis system */

	for ( l = X; l <= Z; l++ )
		for ( m = X; m <= Z; m++ )
		{
			dg[ l ][ m ]  = g_diag[ l ][ m ];
			dgt[ l ][ m ] = g_diag[ m ][ l ];
		}

	for ( l = X; l <= Z; l++ )
	{
		for ( m = X; m <= Z; m++ )
			g_diag[ l ][ m ] = 0.0;
		g_diag[ l ][ l ] = g_eigen[ l ];
	}

	multiply_matrix( 3, g_diag, dgt , g );
	multiply_matrix( 3, dg, g, dgt );

	for ( l = X; l <= Z; l++ )
		for ( m = X; m <= Z; m++ )
			g[ l ][ m ] = dgt[ l ][ m ];

	return 1;
}


/***************************************************/
/* Deallocates memory used during the calculations */
/***************************************************/


void free_memory( DMATRIX g, DMATRIX g_diag, DMATRIX g1, DMATRIX g2,
				  DMATRIX g3, DMATRIX dg, DMATRIX dgt, ATOM *atom,
				  double *energy, int atoms )
{
    free( g );
    free( g_diag );
    free( g1 );
    free( g2 );
    free( g3 );
	free( dg );
	free( dgt );
}
