/********************************************************************

  $Id: g_rhf.h,v 1.11 2000/11/25 14:38:54 jens Exp $

  This file contains definitions and declarations used with g_rhf

*********************************************************************/


#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "m_alloc.h"        /* definitions for allocation of matrices */
#include "diag.h"           /* definitions of diagonalization routines */
#include "matrix.h"         /* definitions of matrix calculations */


#ifndef DEF_PATH
#define DEF_PATH "./"
#endif



typedef struct              /* structure for atoms */
{
	int atom_no;            /* number of atom in input file */
    int n;                  /* quantum number of valence shell */
    double x,               /* cartesian coordinates (in a. u.) */
           y,
           z;
    double Gss,             /* AM1/PM3/MNDO parametrization values */
	       Gsp,             /* one-center two-electron interactions */
	       Gpp,             /* Gss, Gsp, Gpp, Gp2 and Hsp (in a.u.) */
	       Gp2,
	       Hsp,
		   D1,
           D2,              /* D1, D2, rho_0, rho_1 and rho_2 */
           rho_0,           /* (all values in a. u.) */
           rho_1,
           rho_2,
           zeta_s,          /* Slater s- and p-orbital exponents */
           zeta_p,
           sigma_s,         /* < Z / r > in a. u.'s */
           sigma_p,         /* {for Slater s- and p-orbitals) */
           xi,              /* spin-orbit coupling parameter*/
	       tc_fac;          /* factor, used in the calculation of two-center */
	                        /* terms */
    DMATRIX c;              /* matrices for LCAO-coefficents */
} ATOM;


typedef struct              /* structure for atomic parameters */
{
    double Gss,             /* one-center two-electron interactions */
	       Gsp,
	       Gpp,
	       Gp2,
	       Hsp,
		   D1,              /* distance of dipole charge from nucleus */
           D2,              /* distance of quadrupole charge from nucleus */
           rho_0,           /* monopole-monopole interaction parameter */
           rho_1,           /* dipole-dipole interaction parameter */
           rho_2,           /* quadropole-quadropole interaction parameter */
           zeta_s,          /* s-orbital Slater exponent */
           zeta_p,          /* p-orbital Slater exponent */
		   xi;              /* spin-orbit coupling paramter */
} PARAMS;


typedef struct              /* structure for options */
{
	int dn,                /* flag, set if LCAOs are to be renormalized */
	    dec,               /* flag, if set do energy corrections */
	    le,                /* flag, if set log energies */
	    lp,                /* flag, if set log paramagn. contributions */
	    Stone,             /* flag, if set calculate Stones formula */
	                       /* (overrides `do_energy_correction' !) */
	    Angstl,
	    mc,                /* flag, if set multi center contributions */
	                       /* to orbital momentum are calculated */
		half;
	double max_distance;
	char *infile,
	     *outfile,
	     *param_dir;
	FILE *fp;
} OPTION;


typedef struct            /* structure for number of atoms, states etc. */
{
	int atoms,            /* number of atoms in molecule */
	    elecs,            /* number of electrons in molecule */
	    states,           /* size of basis set */
		occ_alpha,        /* number of occupied alpha orbitals */
	    occ_beta,         /* number of occupied beta orbitals */
        SOMO;             /* number of SOMO to be used in indexing of arrays */
} NUMS;


/* Defines of names of atoms */

#define Hydrogen    0
#define Helium      1
#define Lithium     2
#define Beryllium   3
#define Boron       4
#define Carbon      5
#define Nitrogen    6
#define Oxygen      7
#define Fluorine    8
#define Neon        9
#define Sodium     10
#define Magnesium  11
#define Phosphorus 14
#define Sulfur     15
#define Chlorine   16
#define Zinc       29
#define Selenium   33
#define Bromine    34


/* Defines of reduced charges (from Pople-Beveridge) */


#define Z_H       1.00
#define Z_He      1.70
#define Z_Li      1.30
#define Z_Be      1.95
#define Z_B       2.60
#define Z_C       3.25
#define Z_N       3.90
#define Z_O       4.55
#define Z_F       5.20
#define Z_Ne      5.85
#define Z_Na      2.20
#define Z_Mg      2.85
#define Z_P       4.80
#define Z_Zn      4.35
#define Z_S       5.45
#define Z_Cl      6.10


#define X        0
#define Y        1
#define Z        2

#define S         0                 /* angular momentum states */
#define PX        1
#define PY        2
#define PZ        3


#define g_e       2.002319      /* g-factor of free electron */
#define alpha     0.00729735    /* fine structur constant */
#define R_BOHR    0.529177      /* Bohr's radius (in Angstroem) */
#define au_to_eV  27.2116       /* conversion factor between a.u.'s and eV */

#define MAX_PARAMS 54           /* maximum nuclear charge of of atoms in */
                                /* in the parameterization */



/* DEFAULT_DISTANCE: above this value between the atoms all multi-center
   contributions in the calculation of the orbital momentum will be assumed
   to be zero */

#define DEFAULT_DISTANCE  1000.0 / R_BOHR



/* Declarations of functions from g_util.c */

int open_data_file( int argc, char **argv, PARAMS *params, OPTION *options );
int set_up_g_matrices( DMATRIX *g, DMATRIX *g_diag, DMATRIX *g1, DMATRIX *g2,
					   DMATRIX *g3, DMATRIX *dg, DMATRIX *dgt,
					   double *energy, ATOM *atom, int atoms );
void clear( double *energy, ATOM *atom, int atoms );
double sqr( double x );


/* Declarations of functions from read.c */

int get_all_data( NUMS *n, ATOM **atom, double **energy, PARAMS *params,
				  OPTION *options);


/* Declarations of functions from e_corr.c */

int energy_corrections( double **JI, double **KI, ATOM *atom, double *energy,
						NUMS *n, OPTION *options );
int calc_energy_corr( double *JI, double *KI, ATOM *atom, NUMS *n );
int get_memory_for_corrections( double **JI, double **KI, NUMS *n );
void free_energy_corr_memory( double *JI );


/* Declarations of functions from e_util.c */

void make_oc_ints( double tc_int[ 4 ][ 4 ][ 4 ][ 4 ], ATOM *atom );
void make_integrals_1( double *m_int, ATOM *atom, double RAB, int K, int L );
void make_integrals_2( double *m_int, double tc_int[ 4 ][ 4 ][ 4 ][ 4 ],
					   int nA, int nB );


/* Declarations of functions from paramag.c */

int calculate_paramagnetic_part(  DMATRIX g3, ATOM *atom, double *energy,
								  NUMS *n, OPTION *options );


/* Declarations of functions from param_u3.c */

int prep_mc( double *R1, DMATRIX *R2, double *R3, DMATRIX *R4, ATOM *atom,
			 int atoms, int i, int j, OPTION *options );


/* Declarations of functions from relativ.c */

void calculate_relativistic_part( DMATRIX g1, ATOM *atom, NUMS *n );


/* Declarations of functions from diamag.c */

void calculate_diamagnetic_part( DMATRIX g2, ATOM *atom, NUMS *n );
void calculate_diamagnetic_part_Angstl( DMATRIX g2, ATOM *atom, NUMS *n );


/* Declarations of functions from log.c */

void write_results( int diag_ok, DMATRIX g, DMATRIX g_diag, double *g_eigen,
					DMATRIX dg, DMATRIX g1, DMATRIX g2, DMATRIX g3,
					OPTION *options, clock_t starttime );
void log_energy_corrections( char *outfile, double *energy, double *JI,
							 double *K, NUMS *n, OPTION *options );
void log_energy( char *outfile, double *energy, NUMS *n, OPTION *options );
FILE *open_log_file( OPTION *options );
