/***************************************************************************

  $Id: g_util.c,v 1.14 2000/11/09 19:48:21 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.


  Program for calculation of G tensors from RHF data resulting from
  an AM1, PM3, MNDO or ZINDO/S RHF MO calculation using AMPAC or MOPAC
  or Hyperchem/Martin Platos addon (for ZINDO/S). Result is the the G
  tensor in the axis system the LCAO coefficients were calculated in
  as well as the eigenvalues and normalized eigenvectors.

  This file contains some utility routines of the program for the
  calculation of g-tensors.

***************************************************************************/


#include "g_rhf.h"          /* definitions and declarations */



/* Declaration of routines local to this file */

static int check_comand_line( int argc, char **argv, OPTION *options );
static int get_params( PARAMS *params, char *file, OPTION *options );
static int command_error( void );


/*************************************************************************/
/* Tries to open the input data file, reads the type of parametrization, */
/* reads in all needed parameters and returns pointer to data file       */
/*  ->                                                                   */
/*     1. number of command line arguments                               */
/*     2. pointer to command line arguments                              */
/*     3. array of structures for atomic parameter                       */
/*     4. pointer to structure with options                              */
/*  <- 1: everything ok, 0 error detected                                */
/*************************************************************************/

int open_data_file( int argc, char **argv, PARAMS *params, OPTION *options )
{
	FILE *fp;
	char *param_file,
		 *command;


	/* Check command line for flags and file names */

	if ( ! check_comand_line( argc, argv, options ) )
		return 0;

	/* Try to read parameterization parameters */

	if ( ( param_file = malloc( strlen( options->param_dir ) + 129 ) )
		 == NULL )
	{
		printf( "Running out of memory!\n" );
		return 0;
	}
		
	if ( ( command = malloc( strlen( options->param_dir ) +
							 strlen( options->infile ) + 20 ) ) == NULL )
	{
		free( param_file );
		printf( "Running out of memory!\n" );
		return 0;
	}

	sprintf( command, "%sparam.pl %s",
			 options->param_dir, options->infile );

	if ( ( fp = popen( command, "r" ) ) == NULL )
    {
		free( param_file );
		free( command );
        printf( "Can't figure out parameterization used in calculation!\n" );
        return 0;
    }

	strcpy( param_file, options->param_dir );
	if ( fscanf( fp, "%128s", param_file + strlen( param_file ) ) != 1 )
    {
		free( param_file );
		free( command );
        printf( "Can't figure out parameterization used in calculation!\n" );
        return 0;
    }

	free( command );
	pclose( fp );

    /* Read in the parameters for the parameterization to use */

    if ( ! get_params( params, param_file, options ) )
    {
		free( param_file );
        return 0;
    }

	free( param_file );
	return 1;
}


/****************************************************/
/* Interprets the flags on the command line         */
/*  ->                                              */
/*     1. number of command line arguments          */
/*     2. pointer to command line arguments         */
/*     3. pointer to structure of options           */
/*  <- 1 : everything ok, 0 : error in command line */
/****************************************************/

static int check_comand_line( int argc, char **argv, OPTION *options )
{
	int i = 0,
	    mc_t = 0,
	    mc_m = 0;


	/* Clear all flags */

	options->max_distance = -1.0;
	options->dn = options->dec = options->le = options->lp =
	options->Stone = options->Angstl = options->half = options->mc = 0;
	options->param_dir = NULL;

	if ( argc < 2 )             /* if there are no command line options... */
		return command_error( );

	while ( argv[ ++i ][ 0 ] == '-' )
	{
		switch ( argv[ i ][ 1 ] )
		{
			case 'e':         /* flag: Use energy corrections */
				options->dec = 1;
				break;

			case 'D':        /* set integration radius */
				if ( argv[ i ][ 2 ] == '\0' )
				{
					if ( sscanf( argv[ ++i ], "%lf", &options->max_distance )
						 != 1 )
						return command_error( );
				}
				else
				{
					if ( sscanf( argv[ i ] + 2, "%lf", &options->max_distance )
						 != 1 )
						return command_error( );
				}
				options->max_distance /= R_BOHR;
				break;

		    case 'S':         /* use Stone's formulae */
				options->Stone = 1;
				break;

		    case 'w':         /* use Stone's formulae */
				options->half = 1;
				break;

		    case 'A':         /* use Angstl's gauge correction formula */
				options->Angstl = 1;
				break;

			case 't':         /* calculate two-center orbital momenta */
				mc_t = 1;
				break;

			case 'm':         /* calculate multi-center orbital momenta */
				mc_m = 2;
				break;

			case 'l':         /* flag: log energy corr. or paramag. contrib. */
				if ( argv[ i ][ 2 ] == 'e')
				{
				    options->le = 1;
					break;
				}
				if ( argv[ i ][ 2 ] == 'p' )
				{
					options->lp = 1;
					break;
				}

			default:
				printf( "Unrecognized flag: %s\n", argv[ i ] );
				return command_error( );
		}
	}

	if ( options->max_distance < 0.0 )
		options->max_distance = DEFAULT_DISTANCE;

	if ( mc_t )        /* pick right type of contributions to use */
		options->mc = 1;
	if ( mc_m )
		options->mc = 2;

	if ( argc <= i )
		return command_error( );
	else
		options->infile = argv[ i++ ];

	if ( argc <= i )
		options->outfile = NULL;
	else
		options->outfile = argv[ i ];

	/* Figure out where the awk scripts and other auxiliary are located */

	if ( ( options->param_dir = malloc( strlen( DEF_PATH ) + 2 ) ) == NULL )
	{
		printf( "Running out of memory!\n" );
		return 0;
	}

	strcpy( options->param_dir, DEF_PATH );
	if ( options->param_dir[ strlen( options->param_dir ) - 1 ] != '/' )
		strcat( options->param_dir, "/" );

	return 1;
}


/**********************************************************************/
/* Prints a message about the use of the program and always returns 0 */
/**********************************************************************/

static int command_error( void )
{
	printf( "Usage: g_rhf [OPTIONS] FILE [result file]\n"
			" OPTIONS:\n"
			"   -S      : Use Stone's formula\n"
			"   -t      : Use two-center contributions to orbital momentum\n"
			"   -m      : Use multi-center contributions to orbital momentum\n"
			"   -D <nr> : Set maximum distance to <nr> (in A) for multi-center"
			" integrals\n"
			"   -e      : Use energy correction\n"
			"   -w      : Use only half of paramagnetic contribution\n"
			"   -A      : Use Angstl's formula for the gauge correction\n"
			"   -le     : Log energy corrections\n"
			"   -lp     : Log paramagnetic contributions (needs result file)\n"
			"   -h      : Prints this help text and exits\n" );

	return 0;
}


/*********************************************************************/
/* Allocates and initializes memory for all G tensor matrices needed */
/*  ->                                                               */
/*     1. pointer to total g matrix                                  */
/*     2. pointer to diagonalized g matrix                           */
/*     3. pointer to relativistic part of g matrix                   */
/*     4. pointer to diamagnetic part of g matrix                    */
/*     5. pointer to paramagnetic part of g matrix                   */
/*     6. array of alpha energies                                    */
/*     7. array of structures for all atoms                          */
/*     8. number of atoms                                            */
/*  <- 1 : everything ok, 0 : error detected                         */
/*********************************************************************/

int set_up_g_matrices( DMATRIX *g, DMATRIX *g_diag, DMATRIX *g1, DMATRIX *g2,
					   DMATRIX *g3, DMATRIX *dg, DMATRIX *dgt,
					   double *energy, ATOM *atom, int atoms )
{
	int l, m;


	/* Allocate memory for the G tensor matrices */

    *g      = dmatrix( 3, 3 );
    *g_diag = dmatrix( 3, 3 );
    *g1     = dmatrix( 3, 3 );
    *g2     = dmatrix( 3, 3 );
    *g3     = dmatrix( 3, 3 );
	*dg     = dmatrix( 3, 3 );
	*dgt    = dmatrix( 3, 3 );

    if ( ( *g == NULL ) || ( *g_diag == NULL )        /* Allocation failed ! */
        || ( *g1 == NULL ) || ( *g2 == NULL ) || ( *g3 == NULL )
		|| ( *dg == NULL ) || ( *dgt == NULL ) )
    {
        printf( "Not enough memory for G tensor matrices!\n" );
        if ( g != NULL )
            free( *g );
        if ( *g_diag != NULL )
            free( *g_diag );
        if ( *g1 != NULL )
            free( *g1 );
        if ( *g2 != NULL )
            free( *g2 );
        if ( *g3 == NULL )
            free( *g3 );
        if ( *dg != NULL )
            free( *g2 );
        if ( *dgt == NULL )
            free( *g3 );
        clear( energy, atom, atoms );
        return 0;
    }

	/* Initialize G tensor matrices */

    for ( l = X; l <= Z; l++ )
        for ( m = X; m <= Z; m++ )
            ( *g )[ l ][ m ]  = ( *g_diag )[ l ][ m ] =
			( *g1 )[ l ][ m ] = ( *g2 )[ l ][ m ] = ( *g3 )[ l ][ m ] = 0.0;

    ( *g )[ X ][ X ] = ( *g )[ Y ][ Y ] = ( *g )[ Z ][ Z ] = g_e;

	return 1;
}


/**********************************************************/
/* Reads in the parameters for the parametrization used.  */
/*                                                        */
/* Input parameter : File pointer                         */
/*                   Pointer to structures for parameters */
/*                                                        */
/* Return value    : 1 everything ok, 0 error detected    */
/**********************************************************/

static int get_params( PARAMS *params, char *file, OPTION *options )
{
    int i, z, flag;
    double Gss, Gsp, Gpp, Gp2, Hsp,
	       D1, D2, rho_0, rho_1, rho_2, zeta_s, zeta_p, xi;
    char type[ 20 ];
	char *command;
	FILE *fp;


    /* Clear all entries in the array of structures for parameters */

    for ( i = 0; i < MAX_PARAMS; i++ )
    {
		params[ i ].Gss = params[ i ].Gsp = params[ i ].Gpp
			            = params[ i ].Gp2 = params[ i ].Hsp = 0.0;
        params[ i ].D1 = params[ i ].D2 = 0.0;
        params[ i ].rho_0 = params[ i ].rho_1 = params[ i ].rho_2 = 0.0;
        params[ i ].zeta_s = params[ i ].zeta_p = 0.0;
    }
	if ( ( command = malloc( strlen( options->param_dir )
							 + strlen( file ) + 20 ) ) == NULL )
	{
		printf( "Running out of memory!\n" );
		return 0;
	}

	sprintf( command, "%sparams.pl %s", options->param_dir, file );

	if ( ( fp = popen( command, "r" ) ) == NULL )
    {
		free( command );
        printf( "Can't read in parameterization data from `%s'!\n", file );
        return 0;
    }

	free( command );

	if ( fscanf( fp, "%20s", type ) != 1 )
    {
		pclose( fp );
        printf( "Unexpected end of parameter file `%s'!\n", file );
        return 0;
    }

    if ( strcmp( type, "AM1" ) &&
		 strcmp( type, "PM3" ) &&
		 strcmp( type, "MNDO" ) &&
		 strcmp( type, "ZINDOS" ) )
    {
		pclose( fp );
        printf( "Unknown type of parameter file `%s'!\n", file );
        return 0;
    }

    /* Never ever use energy corrections with ZINDO/S calculations ! */

	if ( ! strcmp( type, "ZINDOS" ) )
		 options->dec = 0;

    /* Take care : In the parameter files for AM1 and PM3 from the HYPERCHEM
       program, the definition of rho_0, rho_1 and rho_2 follows the use in
       J.J.P. Stewart, J. Comp. Chem. 10, (1989), 209, while in this program
       the convention from M.J.S. Dewar, W. Thiel, Theoret. Chim. Acta 46,
       (1977), 89, is used. So for the case of AM1 and PM3 the rh0_0, rho_1
       and rho_2 have to be calculated from the values rho_0', rho_1' and
       rho_2' read from the parameter file by
                    rho = 1 / (2 * rho')
       So a flag is set here for AM1 and PM3 parameter files. */

	flag = ! ( strcmp( type, "AM1" ) || strcmp( type, "PM3" ) );

    /* Now read in the parameters */

    while ( fscanf( fp, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
					&z, &Gss, &Gsp, &Gpp, &Gp2, &Hsp, &D1, &D2,
					&rho_0, &rho_1, &rho_2, &zeta_s, &zeta_p ) == 13 )
    {
        if ( z < 1 || z > MAX_PARAMS )
		{
			pclose( fp );
			printf( "Can't handle atoms with charge %d!\n", z );
			return 0;
		}

        params[ z - 1 ].Gss = Gss / au_to_eV;
        params[ z - 1 ].Gsp = Gsp / au_to_eV;
        params[ z - 1 ].Gpp = Gpp / au_to_eV;
        params[ z - 1 ].Gp2 = Gp2 / au_to_eV;
        params[ z - 1 ].Hsp = Hsp / au_to_eV;

        params[ z - 1 ].D1 = D1;
        params[ z - 1 ].D2 = D2;

        if ( flag == 0 )               /* for MNDO parameter file */
        {
            params[ z - 1 ].rho_0 = rho_0;
            params[ z - 1 ].rho_1 = rho_1;
            params[ z - 1 ].rho_2 = rho_2;
        }
        else                         /* for AM1 and PM3 parameter files */
        {
            params[ z - 1 ].rho_0 = 0.5 / rho_0;
            params[ z - 1 ].rho_1 = 0.5 / rho_1;
            params[ z - 1 ].rho_2 = 0.5 / rho_2;
        }

        params[ z - 1 ].zeta_s = zeta_s;
        params[ z - 1 ].zeta_p = zeta_p;
    }

    if ( z < 8 )              /* we expect parameters at least up to oxygen */
    {
		pclose( fp );
        printf( "Unexpected end of parameter file `%s'!\n", file );
        return 0;
    }

	pclose( fp );

	/* Finally read the spin_orbit coupling parameters */

	if ( ( command = malloc( 2 * strlen( options->param_dir ) + 50 ) )
		 == NULL )
	{
		printf( "Running out of memory!\n" );
		return 0;
	}

	sprintf( command, "%sspin_orbit.pl %sspin_orbit",
			 options->param_dir, options->param_dir );

	if ( ( fp = popen( command, "r" ) ) == NULL )
    {
		free( command );
        printf( "Can't read spin-orbit coupling data!\n" );
        return 0;
    }

    while ( fscanf( fp, "%d %lf\n", &z, &xi ) == 2 )
    {
        if ( z < 1 || z > MAX_PARAMS )
		{
			pclose( fp );
			printf( "Can't handle atoms with charge %d!\n", z );
			return 0;
		}

		params[ z - 1 ].xi = xi;
	}

    if ( z < 8 )              /* we expect parameters at least up to oxygen */
    {
		pclose( fp );
        printf( "Unexpected file with spin-orbit couplings!\n" );
        return 0;
    }

	free( command );

    /* Return succesfully */

    return 1;
}


/***********************************************************************/
/* Deallocates memory for energies and matrices with LCAO coefficients */
/* in the structures for the atoms.                                    */
/*                                                                     */
/* Input parameter : Pointer to array of energies                      */
/*                   Pointer to structures of all atoms                */
/*                   Number of atoms                                   */
/***********************************************************************/

void clear( double *energy, ATOM *atom, int atoms )
{
    int i;             /* counter */


    for ( i = 0; i < atoms ; i++ )
        free( atom[ i ].c );

    free( atom );
    free( energy );
}


/**********************************************************************/
/* Calculates the square of a double precision floating point number. */
/*                                                                    */
/* Input parameter : Number to be squared                             */
/* Return value    : Square of number                                 */
/**********************************************************************/

double sqr( double x )
{
    return x * x;
}
