/********************************************************************

  $Id: log.c,v 1.3 2000/11/03 19:08:59 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  This file contains routines for writing results to output files.

********************************************************************/


#include "g_rhf.h"



/************************************************/
/* Writes results of calculation to output file */
/************************************************/

void write_results( int diag_ok, DMATRIX g, DMATRIX g_diag, double *g_eigen,
					DMATRIX dg, DMATRIX g1, DMATRIX g2, DMATRIX g3,
					OPTION *options, clock_t start_time )
{
	FILE *fp;
	int l, m;
	clock_t ct;


    /* Open output file, if no name is given write to standard output */

    if ( options->outfile == NULL )
        fp = stdout;
    else
        if ( ( fp = fopen( options->outfile, "w" ) ) == NULL )
        {
            printf( "Can't open output file: %s !\n", options->outfile );
            fp = stdout;
        }

	/* Write name of input data file to output file */

	fprintf( fp, "\n     Input data file: %s\n", options->infile );

	if ( options->dn )
		fprintf( fp, "     Using renormalized LCAO coefficients\n" );
	if ( options->dec )
		fprintf( fp, "     Using energy corrections\n");
	if (options->mc == 1)
		fprintf( fp, "     Using two-center contributions to orbital "
				 "momentum\n" );
	if ( options->mc == 2 )
		fprintf( fp, "     Using multi-center contributions to orbital "
				 "momentum\n" );
	if ( options->mc == 2 && options->max_distance != DEFAULT_DISTANCE )
		fprintf( fp, "     Using maximum distance for multi-center integrals "
				 "of %.2f A\n", options->max_distance * R_BOHR );
	if ( options->Angstl )
		fprintf( fp, "     Using Angstl's formula for gauge correction\n" );


    /* Now write the 3 types of contributions to the G tensors to the file */

    fprintf( fp, "\n\n     Relativistic correction :\n" );

    for ( l = X; l <= Z; l++ )
    {
        fprintf( fp, "\n" );
        for ( m = 0; m < 3; m++ )
            fprintf( fp, "        % 11.6f", g1[ l ][ m ] );
    }

    fprintf( fp, "\n\n\n     Contribution from gauge correction :\n" );

    for ( l = X; l <= Z; l++ )
    {
        fprintf( fp, "\n" );
        for ( m = X; m <= Z; m++ )
            fprintf( fp, "        % 11.6f", g2[ l ][ m ] );
    }

    fprintf( fp, "\n\n\n     Paramagnetic contribution :\n" );

    for ( l = X; l <= Z; l++ )
    {
        fprintf( fp, "\n" );
        for ( m = X; m <= Z; m++ )
            fprintf( fp, "        % 11.6f", g3[ l ][ m ] );
    }

	if ( diag_ok )
	{
	    fprintf( fp, "\n\n\n     G tensor :\n" );

    	for ( l = X; l <= Z; l++ )
    	{
        	fprintf( fp, "\n" );
	        for ( m = X; m <= Z; m++ )
    	        fprintf( fp, "        % 11.6f", g[ l ][ m ] );
        }

	/* Print eigenvalues and eigenvectors */

		fprintf( fp, "\n\n\n     Eigenvalues :\n\n" );
		for ( l = X; l <= Z; l++ )
			fprintf( fp, "        % 11.6f", g_eigen[ l ] );

		fprintf( fp, "\n\n\n     Eigenvectors (as columns):\n" );
		for ( l = X; l <= Z; l++ )
		{
			fprintf( fp, "\n" );
			for ( m = X; m <= Z; m++ )
				fprintf( fp, "        % 11.6f", dg[ l ][ m ] );
		}
	}

	fprintf( fp, "\n\n" );

	/* Finally add computation time */

	if ( start_time != -1 && ( ct = clock( ) ) != -1 )
		fprintf( fp, "         Computation time:  %.2f seconds\n",
				 ( double ) ( ct - start_time ) / ( double ) CLOCKS_PER_SEC );

    /* Close output file */

    if ( fp != stdout )
        fclose( fp );
}


/************************************************/
/* Writes results of energy corrections to file */
/************************************************/

void log_energy_corrections( char *outfile, double *energy, double *JI,
							 double *KI, NUMS *n, OPTION *options )
{
	FILE *fp;
	int i, j;
	char e_file[ 256 ];


	if ( options->outfile == NULL || options->outfile[ 0 ] == '\0' )
		fp = stdout;
	else
	{
		strcpy( e_file, outfile );
		strcat( e_file, ".energy" );
		if ( ( fp = fopen( e_file, "w" ) ) == NULL )
		{
			printf( "Can't log energy corrections!" );
			return;
		}
	}

	fprintf( fp, "Excitation energies (in eV) into the SOMO:\n\n" );

	for ( i = 0; i < n->occ_beta; i++ )
		fprintf( fp, "%2d->%2d: %7.3f\n", i + 1, n->occ_alpha, au_to_eV *
				 (   energy[ n->SOMO ] - energy[ i ] - JI[ i ] + 0.5 * KI[ i ]
				   + 0.5 * JI[ n->SOMO ] ) );

	fprintf( fp, "\nExcitation energies (in eV) from the SOMO:\n\n" );

	for ( j = n->occ_alpha; j < n->states; j++ )
		fprintf( fp, "%2d->%2d: %7.3f\n", n->occ_alpha, j + 1, au_to_eV *
				(   energy[ j ] - energy[ n->SOMO ] - JI[ j ] + 0.5 * KI[ j ]
				  + 0.5 * JI[ n->SOMO ] ) );

	if ( fp != stdout )
		fclose( fp );
	else
		printf( "\n\n" );
}


/******************************************************************/
/* Writes out excitation energies in case no correctiosn are used */
/******************************************************************/

void log_energy( char *outfile, double *energy, NUMS *n, OPTION *options )
{
	FILE *fp;
	int i, j;
	char e_file[ 256 ];


	if ( options->outfile == NULL || options->outfile[ 0 ] == '\0' )
		fp = stdout;
	else
	{
		strcpy( e_file, outfile );
		strcat( e_file, ".energy" );
		if ( ( fp = fopen( e_file, "w" ) ) == NULL )
		{
			printf( "Can't log energy corrections!" );
			return;
		}
	}

	fprintf( fp, "Excitation energies (in eV) into the SOMO:\n\n" );

	for ( i = 0; i < n->occ_beta; i++ )
		fprintf( fp, "%2d->%2d: %7.3f\n", i + 1, n->occ_alpha,
				 au_to_eV * ( energy[ n->SOMO ] - energy[ i ] ) );

	fprintf( fp, "\nExcitation energies (in eV) from the SOMO:\n\n" );

	for ( j = n->occ_alpha; j < n->states; j++ )
		fprintf( fp, "%2d->%2d: %7.3f\n", n->occ_alpha, j + 1,
				 au_to_eV * ( energy[ j ] - energy[ n->SOMO ] ) );

	if ( fp != stdout )
		fclose( fp );
	else
		printf( "\n\n" );
}


/*********************************************************/
/* Opens the log file for the paramagnetic contributions */
/*********************************************************/

FILE *open_log_file( OPTION *options )
{
	FILE *fp;
	char p_file[ 256 ];


	if ( ! options->lp )
		return NULL;

	if ( options->outfile == NULL || options->outfile[ 0 ] == '\0' )
	{
		printf( "No outfile name given, can't log paramagnetic "
				"contributions!\n" );
		return NULL;
	}

	strcpy( p_file, options->outfile );
	strcat( p_file, ".paramag" );
	if ( ( fp = fopen( p_file, "w" ) ) == NULL )
		printf( "Can't log paramagnetic contributions!\n" );
	else
		fprintf(fp, "Paramagnetic contributions by individual excitations:\n"
				    "     Energy (in eV) and actual contribution to g\n");

	return fp;
}
