/**************************************************************************

  $Id: matrix.c,v 1.5 2001/08/19 20:50:21 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  matrix.C contains various routines for matrix calculations,
  especially the solution of systems of linear equations and
  matrix inversion.

  Contents :
   1. multiply_matrix()       Multiplication of two n * n matrices
   2. lr_decomposition()      LR-decomposition of matrix
   3. lr_backsubstitution()   Solves system of linear equations with
  							  already LR-decomposed coefficient matrix
   4. solve_linear_equation() Solves system of linear equations
   5. invert_matrix()         Matrix inversion
   6. determinat()            Determinant of matrix
   7. lr_determinant()        Determinant of already LR-decomposed
  							  matrix

  ATTENTION : Throughout this routines the definition of a matrix
  			  follows the method described in "Numerical Recipes
  			  in C" (compare also text in m_alloc.c)

**************************************************************************/


/**************************************************************************/
/*                                                                        */
/*      #include's                                                        */
/*                                                                        */
/**************************************************************************/


#include "matrix.h"


/**************************************************************************/
/*                                                                        */
/*      Routines                                                          */
/*                                                                        */
/**************************************************************************/


/*------------------------------------------------------------------------*/
/* multiply_matrix()                                                      */
/*  Multiplication of the two n * n matrices A and B, result is n * n     */
/*  matrix C : C = A * B.                                                 */
/*                                                                        */
/*  ->                                                                    */
/*      1.  Dimension n of both matrices                                  */
/*      2.  Pointer to first matrix A                                     */
/*      3.  Pointer to second matrix B                                    */
/*      4.  Pointer to resulting matrix C                                 */
/*                                                                        */
/*  <-                                                                    */
/*      The result of the multiplication of the matrices A and B will be  */
/*      stored in the matrix C.                                           */
/*------------------------------------------------------------------------*/


void multiply_matrix( int n, DMATRIX a, DMATRIX b, DMATRIX c )
{
    int i, j, k;
    double sum;


    for ( i = 0; i < n; i++ )
        for ( j = 0; j < n; j++ )
        {
            for ( sum = 0.0, k = 0; k < n; k++ )
                sum += a[ i ][ k ] * b[ k ][ j ];
            c[ i ][ j ] = sum;
        }
}
                

/*------------------------------------------------------------------------*/
/* lr_decomposition()                                                     */
/*  Decomposition of the n * n matrix A in two triangular matrices L and  */
/*  R. L is lower triangular (i. e. has nonvanishing elements only on the */
/*  diagonal and below), while R is upper triangular. The diagonal ele-   */
/*  ments of L can be assumed to be equal to one. The result will replace */
/*  the matrix A, the elements below the diagonal belonging to L and the  */
/*  remaining elements on the diagonal and above belonging to R.          */
/*  Since pivoting is necessary some rows of L and R may be interchanged. */
/*  The array 'index' records the row permutations for further calcula-   */
/*  tions - in ith element of 'index' the number of the 'real' row of L   */
/*  and R is stored. The variable 'sign' is +1 or -1 depending on whether */
/*  the number of row interchanges was even or odd, respectively.         */
/*                                                                        */
/*  ->                                                                    */
/*      1.  Dimension n of matrix A                                       */
/*      2.  Pointer to the n * n matrix A                                 */
/*      3.  Pointer to the integer array 'index', having n elements,      */
/*          recording the row interchanges due to pivoting                */
/*      4.  Pointer to real number 'sign', being +1 for even and -1 for   */
/*          odd number of row permutations                                */
/*                                                                        */
/*  <-                                                                    */
/*      Result : 1 = OK, 0 = matrix is singular, -1 = error detected      */
/*      The decomposed matrices L and R replace the initial matrix A      */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


int lr_decomposition( int n, DMATRIX a, int *index, double *sign )
{
    int i,                              /* counter for rows    */
        j,                              /* counter for columns */
        k,                              /* another counter     */
        max_element;                    /* used for pivoting   */
    double *scale,                      /* array for scaling   */
            act_scale,                  /* aktual scaling      */
            sum,                        /* sum variable        */
            temp;                       /* temporary variable  */


    /* Allocate memory for scaling and pivoting */

    if ( ( scale = ( double * ) malloc( n * sizeof( double ) ) ) == NULL )
        return -1;                             /* no more memory -> break */

    /* Initialize 'sign'                                                  */

    *sign = 1.0;

    /* Get maximum element of every rows for pivoting and store it's
       reciprocal */

    for ( i = 0; i < n; i++ )
    {
        for ( act_scale = 0.0, j = 0; j < n; j++ )
            if ( ( temp = fabs( a[ i ][ j ] ) ) > act_scale )
                act_scale = temp;

        if ( act_scale == 0.0 )
        {
            free( scale );
            return 0;                      /* matrix is singular -> break */
        }

        scale[ i ] = 1.0 / act_scale;
    }

    /* Loop over all columns */

    for ( j = 0; j < n; j++ )
    {
        /* Calculate all elements above the diagonal */

        for ( i = 0; i < j; i++ )
        {
            for ( sum = a[ i ][ j ], k = 0; k < i; k++ )
                sum -= a[ i ][ k ] * a[ k ][ j ];
            a[ i ][ j ] = sum;
        }

        /* Initialize for search for largest pivot element */

        act_scale = 0.0;
        max_element = j;

        /* Calculate the elements on and below the diagonal and store the
		   largest element - division by diagonal element will be done later */

        for ( i = j; i < n; i++ )
        {
            for ( sum = a[ i ][ j ], k = 0; k < j; k++ )
                sum -= a[ i ][ k ] * a[ k ][ j ];
            a[ i ][ j ] = sum;

            if ( ( temp = scale[ i ] * fabs( sum ) ) > act_scale )
            {
                max_element = i;
                act_scale = temp;
            }
        }

        /* Test if rows got to be interchanged, record this in 'index' and
		   'sign' */

        if ( j != max_element )
        {
            for ( k = 0; k < n; k++ )
            {
                temp = a[ max_element ][ k ];
                a[ max_element ][ k ] = a[ j ][ k ];
                a[ j ][ k ] = temp;
            }

            scale[ max_element ] = scale[ j ];
            *sign *= -1.0;
        }

        index[ j ] = max_element;

        /* Check again for singularity of matrix */

        if ( ( temp = a[ j ][ j ]) == 0.0 )
        {
            free( scale );
            return 0;                      /* matrix is singular -> break */
        }

        /* Finally divide the elements of the current column below the
		   diagonal by the diagonal element */

        temp = 1.0 / temp;
        for ( i = j + 1; i < n; i++ )
            a[ i ][ j ] *= temp;
    }

    /* Deallocate memory and return */

    free( scale );

    return 1;
}


/*------------------------------------------------------------------------*/
/* lr_backsubstitution()                                                  */
/*  Solves a n-dimensional system of linear equations if the coefficent   */
/*  matrix A is already decomposed in the two triangular matrices L and   */
/*  R (cf. the function lr_decomposition()).                              */
/*                                                                        */
/*  ->                                                                    */
/*      1.  Dimension n of the system of linear equations                 */
/*      2.  Pointer to LR-decomposed n * n coefficent matrix              */
/*      3.  Pointer to integer array with n elements, recording the row   */
/*          permutations due to pivoting                                  */
/*      4.  Pointer to right-hand side vector 'b' of the system of linear */
/*          equations A * x = b                                           */
/*                                                                        */
/*  <-                                                                    */
/*      'b' returns the solution vector x                                 */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


void lr_backsubstitution( int n, DMATRIX a, int *index, double *b )
{
    int i, j,                           /* temporary counters */
        start = -1,
        act_index;
    double sum;                         /* sum variable */


    /* First, solve the system of equations L * y = b. Take care of row
	   interchanges due to pivoting. Start summation only when the first
	   nonzero element of 'b' is encountered */

    for ( i = 0; i < n; i++ )
    {
        act_index = index[ i ];
        sum = b[ act_index ];
        b[ act_index ] = b[ i ];

        if ( start != - 1 )
            for ( j = start; j < i; j++ )
                sum -= a[ i ][ j ] * b[ j ];
        else
            if ( sum != 0.0 )
                start = i;

        b[ i ] = sum;
    }

    /* Solve R * x = y to get the final result */

    for ( i = n - 1; i >= 0; i-- )
    {
        sum = b[ i ];
        if ( i < n - 1 )
            for ( j = i + 1; j < n; j++ )
                sum -= a[ i ][ j ] * b[ j ];
        b[ i ] = sum / a[ i ][ i ];
    }
}


/*------------------------------------------------------------------------*/
/* solve_linear_equation()                                                */
/*  Solves a n-dimensional system of linear equations A * x = b by        */
/*  LR-decomposition.                                                     */
/*  First, the coefficent matrix A is decomposed in two triangular ma-    */
/*  trices L and R, then the solution vector x is calculated.             */
/*  To solve the system of equations for a different right-hand side      */
/*  vector b it is unnecessary to repeat the LR-decomposition, instead    */
/*  the function lr_backsubstiution() can be used directly. In this case  */
/*  a pointer to a n-dimensional array has to be passed to this function  */
/*  which is needed later on by the function lr_backsubstiution(). Other- */
/*  wise this pointer has to be zero !                                    */
/*                                                                        */
/*  ->  1.  Dimension n of the system of linear equations                 */
/*      2.  Pointer to the n * n coefficent matrix A                      */
/*      3.  Pointer to the right-hand side vector b                       */
/*      4.  Pointer to n-dimensional integer array 'index', or zero if    */
/*          not needed                                                    */
/*                                                                        */
/*  <=  Result : 1 = OK, 0 : matrix is singular, -1 = error detected      */
/*                                                                        */
/*      'b' returns the solution vector x, 'index' records the permu-     */
/*      tations due to pivoting (if not set to zero initially)            */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


int solve_linear_equation( int n, DMATRIX a, double *b, int *index )
{
    int *temp_index,
        lr_result;
    double sign;


    /* If 'index' is zero, allocate memory for 'index' */

    if ( index == NULL )
    {
        if ( ( temp_index = ( int * ) malloc( n * sizeof( int ) ) ) == NULL )
            return -1;                         /* no more memory -> break */
    }
    else
        temp_index = index;

    /* LR-decompose coefficent matrix A */

    if ( ( lr_result = lr_decomposition( n, a, temp_index, &sign ) ) != 1 )
    {
        if ( index == NULL )
            free( temp_index );
        return lr_result;            /* matrix singular or error -> break */
    }

    /* Calculate solution vector 'b' */

    lr_backsubstitution( n, a, temp_index, b );

    /* If 'index' was zero initially deallocate memory */

    if ( index == NULL )
        free( temp_index );

    return 1;
}


/*------------------------------------------------------------------------*/
/* invert_matrix()                                                        */
/*  Calculate inverse of a n * n matrix A by solving the system of linear */
/*  equations A * x = b with all columns of the identity matrix as right- */
/*  hand side vectors. The inverse of A replaces the initial matrix A.    */
/*                                                                        */
/*  ->                                                                    */
/*      1. Dimension n of the matrix A                                    */
/*      2. Pointer to the matrix A                                        */
/*                                                                        */
/*  <-                                                                    */
/*      Result : 1 = OK, 0 = matrix is singular, -1 = error detected      */
/*      The inverse of A replaces the initial matrix A.                   */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


int invert_matrix( int n, DMATRIX a )
{
    int i, j,
        *index,
        lr_result;
    DMATRIX inverse;
    double sign;


    /* Allocate memory for inverse matrix and array for recording row
	   permutations due to pivoting */

    if ( ( inverse = dmatrix( n, n ) ) == NULL )
        return -1;

    if ( ( index = ( int * ) malloc( n * sizeof( int ) ) ) == NULL )
        return -1;

    /* LR-decompose initial Matrix A */

    if ( ( lr_result = lr_decomposition( n, a, index, &sign ) ) != 1 )
    {
        free( inverse );
        free( index );
        return lr_result;            /* matrix singular or error -> break */
    }

    /* Solve system of linear equations for all columns of the identity
	   matrix */

    for ( i = 0; i < n; i++ )
    {
        for ( j = 0; j < n; j++ )
            inverse[ i ][ j ] = 0.0;
        inverse[ i ][ i ] = 1.0;

        lr_backsubstitution( n, a, index, inverse[ i ] );
    }

    /* Transpose resulting matrix and store in initial matrix A */

    for ( i = 0; i < n; i++ )
    {
        a[ i ][ i ] = inverse[ i ][ i ];
        for ( j = 0; j < i; j++ )
        {
            a[ i ][ j ] = inverse[ j ][ i ];
            a[ j ][ i ] = inverse[ i ][ j ];
        }
    }

    /* Deallocate memory and return */

    free( inverse );
    free( index );

    return 1;
}


/*------------------------------------------------------------------------*/
/* determinat()                                                           */
/*  Calculate determinant of n * n matrix A by LR-decomposition.          */
/*  The initial matrix A will be destroyed !                              */
/*                                                                        */
/*  ->  1.  Dimension n of the matrix A                                   */
/*      2.  Pointer to the matrix A                                       */
/*      3.  Pointer to the result det(A)                                  */
/*                                                                        */
/*  <-                                                                    */
/*      Result : 1 = OK, -1 : error detected                              */
/*      The determinant of A is stored in the variable 'det', the initial */
/*      matrix A will be destroyed.                                       */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


int determinant( int n, DMATRIX a, double *det )
{
    double sign;
    int *index,
        lr_result;


    /* Allocate memory for the array for recording row permutations done in
	   pivoting  */

    if ( ( index = ( int * ) malloc( n * sizeof( int ) ) ) == NULL )
        return -1;                             /* no more memory -> break */

    /* LR-decompose matrix A */

    lr_result = lr_decomposition( n, a, index, &sign );

    /* Deallocate memory */

    free( index );

    /* Test result of decomposition */

    if ( lr_result == -1 )
        return -1;                             /* error detected -> break */

    if ( lr_result == 0 )
    {
        *det = 0.0;                   /* matrix is singular -> det(A) = 0 */
        return 1;
    }

    /* Calculate determinant of LR-decomposed matrix */

    *det = lr_determinant( n, a, sign );

    return 1;
}


/*------------------------------------------------------------------------*/
/* lr_determinant()                                                       */
/*  Calculates determinant of LR-decomposed matrix.                       */
/*                                                                        */
/*  ->                                                                    */
/*      1.  Dimension n of LR-decomposed matrix A                         */
/*      2.  Pointe to LR-decomposed matrix A                              */
/*      3.  real number 'sign' as returned by lr_decomposition(), reflec- */
/*          ting the number of row permutations due to pivoting           */
/*                                                                        */
/*  <-                                                                    */
/*      Result : Determinant of the LR-decomposed matrix                  */
/*                                                                        */
/*  Source : 'Numerical Recipes'                                          */
/*------------------------------------------------------------------------*/


double lr_determinant( int n, DMATRIX a, double sign )
{
    double det;
    int i;


    /* Initialize 'det' and calculate product of diagonal elements */

    for ( det = sign, i = 0; i < n; i++ )
        det *= a[ i ][ i ];

    return det;
}
