#!/usr/local/bin/perl -w
#
# $Id: numbers.pl,v 1.7 2000/11/10 11:41:35 jens Exp $
#
# Copyright (C) 2001 Jens Thoms Toerring
#
# g_rhf is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# g_rhf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with g_rhf; see the file COPYING.  If not, write to
# the Free Software Foundation, 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# Parses a RHF MOPAC or AMPAC or Hyperchem/Plato ZINDO/S result file and the
# number of atoms, electrons, the total number of orbitals and the number of
# occupied beta-orbitals and the number of occupied alpha-orbitals (difference
# should always be 1 because we're dealing with doublet states).


use strict;

local *F;
my $method;
my $num_atoms = 0;
my $num_open;
my $num_filled;
my $num_electrons = 0;
my $num_orbitals;
my $in_coordinates = 0;
my @list;
my %charges = ( 'H' =>   1,
				'HE' =>  2,
				'LI' =>  3,
				'BE' =>  4,
				'B'  =>  5,
				'C'  =>  6,
				'N'  =>  7,
				'O'  =>  8,
				'F'  =>  9,
				'NE' => 10,
				'NA' => 11,
				'MG' => 12,
				'AL' => 13,
				'SI' => 14,
				'P'  => 15,
				'S'  => 16,
				'CL' => 17,
				'AR' => 18,
				'K'  => 19,
				'CA' => 20,
				'SC' => 21,
				'TI' => 22,
				'V'  => 23,
				'CR' => 24,
				'MN' => 25,
				'FE' => 26,
				'CO' => 27,
				'NI' => 28,
				'CU' => 29,
				'ZN' => 30,
				'GA' => 31,
				'GE' => 32,
				'AS' => 33,
				'SE' => 34,
				'BR' => 35,
				'CD' => 48,
				'IN' => 49,
				'SN' => 50,
				'SB' => 51,
				'TE' => 52,
				'I'  => 53 );


open( F, "< $ARGV[ 0 ]" ) or exit;

while ( <F> ) {

	if ( /\b((MOPAC)|(AMPAC)|(ZINDOS))\b/ ) {
		$method = uc $1;
		last;
	}
}

exit unless defined $method;

if ( $method =~ /(AMPAC)|(MOPAC)/ ) {

	while ( <F> ) {
		last if /CARTESIAN COORDINATES/i;
	}

	&{ \&{ $method . "_coordinates" } };

	while ( <F> ) {
		next unless /NO\. OF FILLED LEVELS/i or /AND NO\. OF OPEN LEVELS/;
		chomp;
		@list = split /\s*=\s*/;
		if ( /NO\. OF FILLED LEVELS/i ) {
			$num_filled = $list[ 1 ];
		} else {
			$num_open = $list[ 1 ];
			last;
		}
	}

	while ( <F> ) {
		next unless /^\s*Root No\./i;
		s/\s+/ /g;
		/^\s*(.*)\b\s*$/;
		@list = split /\s+/, $1;
		$num_orbitals = $list[ $#list ];
	}
} else {                                       # ZINDO/S

	while ( <F> ) {
		last if /NET CHARGES AND COORDINATES/
	}

	&{ \&{ $method . "_coordinates" } };

	while ( <F> ) {
		next unless /Total number of MOs and AOs, Norb/i;
		s/s+/ /g;
		/\s*(.*)\b\s*/;
		@list = split /\s*=\s*/, $1;
		$num_orbitals = $list[ $#list ];
		last;
	}

	while ( <F> ) {
		next unless /Number of half-filled MO/i;
		s/s+/ /g;
		/\s*(.*)\b\s*/;
		@list = split /\s*=\s*/, $1;
		$num_filled = $list[ $#list ] - 1;
		$num_open = 1;
		last;
	}

}

close F;

print "$num_atoms $num_electrons $num_orbitals $num_filled ",
	   $num_filled + $num_open, "\n";


sub MOPAC_coordinates {
	while ( <F> ) {
		next unless /^\s*\d+\s+([A-Z]{1,2}|[+-])\s+/i or $in_coordinates;
		next if /^\s*\d+\s+([+-])|(xx)\s+/i;
		last unless /^\s*\d+\s+([A-Z]{1,2}|[+-])\s+/i;
		$in_coordinates = 1;
		s/\s+/ /g;
		/^\s*(.*)\b\s*$/;
		@list = split /\s/, $1;
		$num_atoms++;
		$num_electrons += $charges{ uc $list[ 1 ] };
	}
}

sub AMPAC_coordinates {
	while ( <F> ) {
		next unless /^\s*\d+\s+\d+\s+/i or $in_coordinates;
		next if /^\s*\d+\s+[+-]\s+/;
		last unless /^\s*\d+\s+(\d+)\s+/i;
		$in_coordinates = 1;
		s/\s+/ /g;
		/^\s*(.*)\b\s*$/;
		@list = split /\s/, $1;
		$num_atoms++;
		$num_electrons += $list[ 1 ];
	}
}

sub ZINDOS_coordinates {
	while ( <F> ) {
		next unless /\s*\d+\s+\d+\s+/ or $in_coordinates;
		last unless /\s*\d+\s+\d+\s+/;
		$in_coordinates = 1;
		s/s+/ /g;
		/\s*(.*)\b\s*/;
		@list = split /\s+/, $1;
		$num_atoms++;
		$num_electrons += $list[ 1 ];
	}
}
