/******************************************************************

  $Id: param_u.c,v 1.7 2000/11/03 23:22:40 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  Here some of the utility routines used in the calculation of
  the paramagnetic part of the g-tensor are defined.

******************************************************************/


#include "g_rhf.h"                /* definitions and declarations */


/* Declaration of routines local to this file */


static void rotate_lcaos( ATOM *L, ATOM *M, int i, double R[ 3 ][ 3 ],
						  double *clM, double RLM );
static void make_aux_ints( double R, double zeta1, double zeta2,
						   double *AI, double *BI );
static void make_A_ints( double p, double *AI );
static void make_B_ints( double q, double *BI );


/***************************************************************************/
/* In this routine most of the work for the calculation of the expectation */
/* value of the orbital momentum including two- and three-center terms is  */
/* done (in the RHF case). Afterwards only the routine mc_lev_s() has to   */
/* to be called to complete the calculation.                               */
/***************************************************************************/


int prep_mc( double *R1, DMATRIX *R2, double *R3, DMATRIX *R4, ATOM *atom,
			 int atoms, int i, int j, OPTION *options )
{
    DMATRIX cL;                   /* matrix of LCAO coefficents on L */
    double clL[ 4 ],              /* LCAO coefficients in local system on */
           clM[ 4 ];              /* atom L and M */
    double R[ 3 ][ 3 ];           /* rotation matrix between main and local */
                                  /* coordinate system */
    double RLM,                   /* distance between atoms L and M */
           RLM3, RLM4, RLM5;      /* and powers of half the distance */
    int L, M;                     /* counter for atoms */
    int k;                        /* one more counter ;-) */
    double AI[ 5 ], BI[ 5 ];      /* auxiliary integrals */
    double sum_x, sum_y, sum_z;
    int a, b, c;                  /* coordinates */
    double zsL  = 0.0,            /* Slater-orbital coefficients of atoms */
		   zpL  = 0.0,            /* L and M and powers of them */
           zsL5 = 0.0,
		   zpL5 = 0.0,
           zsL7 = 0.0,
		   zpL7 = 0.0,
           zsM  = 0.0,
		   zpM  = 0.0,
           zsM3 = 0.0,
		   zsM5 = 0.0,
		   zpM5 = 0.0;

	/* First some memory for the two- and three-center terms is needed */

	if ( options->mc > 0 && *R2 == NULL &&
		 ( *R2 = dmatrix( 3, atoms ) ) == NULL )
	{
		printf( "Not enough memory for calculation of three-center "
				"terms !\n" );
		return 0;
	}

	if ( options->mc == 2 && *R4 == NULL &&
		 ( *R4 = dmatrix( 3, atoms ) ) == NULL )
	{
		free( *R2 );
		printf( "Not enough memory for calculation of three-center "
				"terms !\n");
		return 0;
	}

	for ( a = X; a <= Z; a++ )
	{
		R1[ a ] = R3[ a ] = 0.0;
		for ( L = 0; L < atoms; L++ )
		{
			( *R2 )[ a ][ L ] = 0.0;
			if ( *R4 != NULL )
				( *R4 )[ a ][ L ] = 0.0;
		}
	}

	for ( L = 0; L < atoms; L++ )
	{
		/* Use temporary variable */

		cL = atom[ L ].c;

		/* Calculate some often used values */

		zsL  = atom[ L ].zeta_s;
		zsL5 = pow( zsL, 2.5 );
		zsL7 = pow( zsL, 3.5 );

        /* First we calculate the one- and two-center integrals */

        if ( atom[ L ].n > 1 )
		{
			zpL  = atom[ L ].zeta_p;
			zpL5 = pow( zpL, 2.5 );
			zpL7 = pow( zpL, 3.5 );

			for ( a = X; a <= Z; a++ )
			{
				b = ( a + 1 ) % 3;
				c = ( a + 2 ) % 3;

				R1[ a ] +=   cL[ c + PX ][ i ] * cL[ b + PX ][ j ]
					       - cL[ b + PX ][ i ] * cL[ c + PX ][ j ];

				( *R2 )[ a ][ L ] = atom[ L ].tc_fac *
					                (   cL[ a + PX ][ i ] * cL[ S ][ j ]
									  - cL[ S ][ i ] * cL[ a + PX ][ j ] );
			}
		}

		if ( options->mc < 2 ) /* if no three-center integrals are to be     */
			continue;          /* included we're already finished for atom L */


		/* And now follow the three-center terms */

        for ( M = 0; M < atoms; M++ )
        {
            if ( L == M )                  /* this case is already treated ! */
                continue;

            /* First we calculate values which are often used in this loop */

            RLM = sqrt(   sqr( atom[ M ].x - atom[ L ].x )
                        + sqr( atom[ M ].y - atom[ L ].y )
                        + sqr( atom[ M ].z - atom[ L ].z ) );

			if ( RLM > options->max_distance )      /* atoms too far apart ? */
				continue;                           /* then skip the rest    */

            RLM3 = pow( 0.5 * RLM, 3.0 );
            RLM4 = pow( 0.5 * RLM, 4.0 );
            RLM5 = pow( 0.5 * RLM, 5.0 );

            zsM  = atom[ M ].zeta_s;
            zsM3 = pow( zsM, 1.5 );
            zsM5 = pow( zsM, 2.5 );

            if ( atom[ M ].n > 1 )
            {
                zpM  = atom[ M ].zeta_p;
                zpM5 = pow( zpM, 2.5 );
            }

			/* The LCAO coefficients of atom M have to be "rotated" into the
			   loal coordinate system */

			rotate_lcaos( atom + L, atom + M, i, R, clM, RLM );

            /* Now finally we calculate the three-center integrals ... */

            if ( atom[ L ].n == 1 && atom[ M ].n == 1 )
            {
                make_aux_ints( RLM, zsL, zsM, AI, BI );

                /* < 1S' | d/db | 1S' > */

				for ( a = X; a <= Z; a++ )
					(*R4)[a][L] += 2.0 * clM[S] * cL[S][j] * zsM3 * zsL5 * RLM3
						* R[Z][a] * (   AI[2] * BI[1] + AI[1] * (BI[0] - BI[2])
									  - AI[0] * BI[1]);
            }

            if ( atom[ L ].n == 1 && atom[ M ].n == 2 )
            {
                make_aux_ints( RLM, zsL, zsM, AI, BI );

                /* < 2S' | d/da | 1S' > */

				for ( a = X; a <= Z; a++ )
					(*R4)[a][L] += 2.0 / sqrt(3.0) * clM[S] * cL[S][j]
						* zsM5 * zsL5 * RLM4 * R[Z][a]
						* (   AI[3] * BI[1] + AI[2] * (BI[0] - 2.0 * BI[2])
						    + AI[1] * (BI[3] - 2.0 * BI[1]) + AI[0] * BI[2]);

                make_aux_ints( RLM, zsL, zpM, AI, BI );

                /* < 2P' | d/da | 1S'> */

				for ( a = X; a <= Z; a++ )
					(*R4)[a][L] += cL[S][j] * zpM5 * zsL5 * RLM4
						     * (   (R[X][a] * clM[PX] + R[Y][a] * clM[PY])
								 * (   (AI[3] - AI[1]) * (BI[0] - BI[2])
								     - (AI[2] - AI[0]) * (BI[1] - BI[3]))
								 + 2.0 * clM[PZ] * R[Z][a]
								 * ( - AI[3] * BI[2] + AI[2] * BI[3]
									 + AI[1] * BI[0] - AI[0] * BI[1]));
            }

            if ( atom[ L ].n == 2 && atom[ M ].n == 1 )
            {
                sum_x = R[X][X] * cL[PX][j] + R[X][Y] * cL[PY][j]
                                            + R[X][Z] * cL[PZ][j];
                sum_y = R[Y][X] * cL[PX][j] + R[Y][Y] * cL[PY][j]
                                            + R[Y][Z] * cL[PZ][j];
                sum_z = R[Z][X] * cL[PX][j] + R[Z][Y] * cL[PY][j]
                                            + R[Z][Z] * cL[PZ][j];

                make_aux_ints( RLM, zsL, zsM, AI, BI );

				for ( a = X; a <= Z; a++ )
				{
					/* < 1S' | T1' > */

					(*R4)[a][L] -= 2.0 / sqrt(3.0) * clM[S] * cL[S][j]
						    * zsM3 * zsL5 * RLM3 *R[Z][a] 
							* (   AI[2] * BI[1] + AI[1] * (BI[0] - BI[2])
							    - AI[0] * BI[1]);

					/* < 1S' | T2' > */

					(*R4)[a][L] += 2.0 / sqrt(3.0) * clM[S] * cL[S][j]
						    * zsM3 * zsL7 * RLM4 * R[Z][a]
							* (   AI[3] * BI[1] + AI[2] * BI[0] - AI[1] * BI[3]
					            - AI[0] * BI[2]);
				}

                make_aux_ints( RLM, zpL, zsM, AI, BI );

                /* < 1S' | 2Pz' > (S_3) */

				for ( a = X; a <= Z; a++ )
				{
					b = ( a + 1 ) % 3;
					c = ( a + 2 ) % 3;
					clL[ PZ ] =   cL[ b + PX ][ j ] * R[ Z ][ c ]
						        - cL[ c + PX ][ j ] * R[ Z ][ b ];

					/* < 1S' | d/da | 2Pz' > (S_3) */

					R3[a] += 2.0 * clM[S] * clL[PZ] * zsM3 * zpL5 * RLM4
						    * (   AI[3] * BI[1] + AI[2] * BI[0] - AI[1] * BI[3]
							    - AI[0] * BI[2]);

					/* < 1S' | T3'> */

					(*R4)[a][L] -= 2.0 * clM[S] * zpL5 * zsM3 * RLM3
						     * cL[a + PX][j] * (AI[2] * BI[0] - AI[0] * BI[2]);


					/* < 1S' | T4' > */

					(*R4)[a][L] += clM[S] * zpL7 * zsM3 * RLM4
						* (   (R[X][a] * sum_x + R[Y][a] * sum_y)
						    * (   (AI[3] - AI[1]) * (BI[0] - BI[2])
							   - (AI[2] - AI[0]) * (BI[1] - BI[3]))
						    + 2.0 * sum_z * R[Z][a]
						    * (   AI[3] * BI[2] - AI[2] * (BI[3] - 2.0 * BI[1])
							    + AI[1] * (BI[0] - 2.0 * BI[2])
							    - AI[0] * BI[1]));
				}
			}

            if ( atom[ L ].n == 2 && atom[ M ].n == 2 )
            {
                sum_x = R[X][X] * cL[PX][j] + R[X][Y] * cL[PY][j]
                                            + R[X][Z] * cL[PZ][j];
                sum_y = R[Y][X] * cL[PX][j] + R[Y][Y] * cL[PY][j]
                                            + R[Y][Z] * cL[PZ][j];
                sum_z = R[Z][X] * cL[PX][j] + R[Z][Y] * cL[PY][j]
                                            + R[Z][Z] * cL[PZ][j];

                make_aux_ints( RLM, zsL, zsM, AI, BI );

				for ( a = X; a <= Z; a++ )
				{
					/* < 2S' | T1' > */

					(*R4)[a][L] -= 2.0 / 3.0 * clM[S] * cL[S][j]
						  * zsL5 * zsM5 * RLM4 * R[Z][a]
						  * (   AI[3] * BI[1] + AI[2] * (BI[0] - 2.0 * BI[2])
							  - AI[1] * (2.0 * BI[1] - BI[3]) + AI[0] * BI[2]);

					/* < 2S' | T2' > */

					(*R4)[a][L] += 2.0 / 3.0 * clM[S] * cL[S][j]
						       * zsL7 * zsM5 * RLM5 * R[Z][a]
							   * (   AI[4] * BI[1] + AI[3] * (BI[0] - BI[2])
							       - AI[2] * (BI[1] + BI[3])
							       + AI[1] * (BI[4] - BI[2]) + AI[0] * BI[3]);
				}

                make_aux_ints( RLM, zpL, zsM, AI, BI );

				for ( a = X; a <= Z; a++ )
				{
					b = ( a + 1 ) % 3;
					c = ( a + 2 ) % 3;
                    clL[ PZ ] =   cL[ b + PX ][ j ] * R[ Z ][ c ]
						        - cL[ c + PX ][ j ] * R[ Z ][ b ];

					/* < 2S' | 2Pz' > (S_3) */

					R3[a] += 2.0 / sqrt(3.0) * clM[S] * clL[PZ]
						    * zsM5 * zpL5 * RLM5
							* (   AI[4] * BI[1] + AI[3] * (BI[0] - BI[2])
							    - AI[2] * (BI[1] + BI[3])
							    + AI[1] * (BI[4] - BI[2]) + AI[0] * BI[3]);

					/* < 2S' | T3' > */

					(*R4)[a][L] -= 2.0 / sqrt(3.0) * clM[S]
						         * zpL5 * zsM5 * RLM4 * cL[a + PX][j]
								 * (   AI[3] * BI[0] - AI[2] * BI[1]
									 - AI[1] * BI[2] + AI[0] * BI[3]);

					/* < 2S' | T4' > */

					(*R4)[a][L] += clM[S] * zpL7 * zsM5 * RLM5 / sqrt(3.0)
						* (   (   R[X][a] * sum_x + R[Y][a] * sum_y)
						    * (   AI[4] * (BI[0] - BI[2])
							    - 2.0 * (AI[3] - AI[1]) * (BI[1] - BI[3])
							    - AI[2] * (BI[0] - 2.0 * BI[2] + BI[4])
							    + AI[0] * (BI[4] - BI[2]))
						    + 2.0 * sum_z * R[Z][a]
						    * (   AI[4] * BI[2]
							    - 2.0 * (AI[3] - AI[1]) * (BI[3] - BI[1])
							    + AI[2] * (BI[0] - 4.0 * BI[2] + BI[4])
							    + AI[0] * BI[2]));
				}

                make_aux_ints( RLM, zsL, zpM, AI, BI );

				for ( a = X; a <= Z; a++ )
				{
					/* < 2Pd' | T1' > (d = x, y) */

					(*R4)[a][L] -= cL[S][j] * zsL5 * zpM5 * RLM4 / sqrt(3.0)
						    * (R[X][a] * clM[PX] + R[Y][a] * clM[PY])
							* (   (AI[3] - AI[1]) * (BI[0] - BI[2])
							    - (AI[2] - AI[0]) * (BI[1] - BI[3]));

					/* < 2Pz' | T1'> */

					(*R4)[a][L] -= 2.0 / sqrt(3.0) * cL[S][j] * clM[PZ]
						    * zsL5 * zpM5 * RLM4 * R[Z][a]
							* ( - AI[3] * BI[2] + AI[2] * BI[3] + AI[1] * BI[0]
							    - AI[0] * BI[1]);

					/* < 2Pd' | T2' > (d = x, y) */

					(*R4)[a][L] += cL[S][j] * zsL7 * zpM5 * RLM5 / sqrt(3.0)
						* (R[X][a] * clM[PX] + R[Y][a] * clM[PY])
						* (   AI[4] * (BI[0] - BI[2]) + AI[2] * (BI[4] - BI[0])
						    + AI[0] * (BI[2] - BI[4]));

					/* < 2Pz' | T2' > */

					(*R4)[a][L] += 2.0 / sqrt(3.0) * cL[S][j] * clM[PZ]
						    * zsL7 * zpM5 * RLM5 * R[Z][a]
							* ( - AI[4] * BI[2] + AI[2] * (BI[4] + BI[0])
							    - AI[0] * BI[2]);
				}

                make_aux_ints( RLM, zpL, zpM, AI, BI );

				for ( a = X; a <= Z; a++ )
				{
					b = ( a + 1 ) % 3;
					c = ( a + 2 ) % 3;
					for ( k = X; k <= Z; k++ )
						clL[ k + PX ] =   cL[ b + PX ][ j ] * R[ k ][ c ]
							            - cL[ c + PX ][ j ] * R[ k ][ b ];

					/* < 2Pd' | 2Pd' > (d = x, y) (S_3) */

					R3[a] += (clM[PX] * clL[PX] + clM[PY] * clL[PY])
						* zpM5 * zpL5 * RLM5
						* (   AI[4] * (BI[0] - BI[2]) + AI[2] * (BI[4] - BI[0])
						    + AI[0] * (BI[2] - BI[4]));

					/* < 2Pz' | 2Pz' > (S_3) */

					R3[a] += 2.0 * clM[PZ] * clL[PZ] * zpM5 * zpL5 * RLM5
						     * ( - AI[4] * BI[2] + AI[2] * (BI[4] + BI[0])
								 - AI[0] * BI[2]);

					/* < 2Pz' | T3' > */

					(*R4)[a][L] -= 2.0 * clM[PZ] * zpL5 * zpM5 * RLM4
						    * cL[a + PX][j]
						    * ( - AI[3] * BI[1] + AI[2] * BI[0] + AI[1] * BI[3]
							    - AI[0] * BI[2]);

					/* < 2Pd' | T4' > (d = x, y) */

					(*R4)[a][L] += zpL7 * zpM5 * RLM5
                        * (  clM[PX] * (R[Z][a] * sum_x + R[X][a] * sum_z)
                           + clM[PY] * (R[Z][a] * sum_y + R[Y][a] * sum_z))
						* (   (AI[4] - 2.0 * AI[2] + AI[0]) * (BI[1] - BI[3])
						    + (AI[3] - AI[1]) * (BI[4] - 2.0 * BI[2] + BI[0]));

					/* < 2Pz' | T4' > */

					(*R4)[a][L] += clM[PZ] * zpL7 * zpM5 * RLM5
                    * (   (R[X][a] * sum_x + R[Y][a] * sum_y)
                        * (   (AI[4] - AI[0]) * (BI[3] - BI[1])
						    + (AI[3] - AI[1]) * (BI[0] - BI[4]))
                        + 2.0 * sum_z * R[Z][a]
                        * ( - AI[4] * BI[3] + AI[3] * (BI[4] - BI[2])
						    + AI[2] * (BI[3] + BI[1])
                            + AI[1] * (BI[0] - BI[2]) - AI[0] * BI[1]));
				}
            }

        }  /* End of loop over all atoms M */
    } /* End of loop over all atoms L */

	return 1;
}


/************************************************************************/
/* Rotates the CAO coefficient on atom M into a local coordinate system */
/* where the z-axis connects the atoms L and M.                         */
/*  ->                                                                  */
/*     1. pointer to structure of atom L                                */
/*     2. pointer to structure of atom M                                */
/*     3. number of orbital                                             */
/*     4. rotation matrix                                               */
/*     5. LCAO on M in local system                                     */
/*     6. distance between atom L and M                                 */
/************************************************************************/

static void rotate_lcaos( ATOM *L, ATOM *M, int i, double R[ 3 ][ 3 ],
						  double *clM, double RLM )
{
    double cp, sp, ct, st,        /* sines and cosines of rotation angles */
           temp;                  /* between main and local system */


	/* The two-center integrals have to be calculated in a local coordinate
	   system, were the z-axis connects the atoms L and M. We get to this
	   local system by rotating first around the z-axis by an angle phi,
	   followed by a rotation  around the new y'-axis by an angle theta.
	   We do this by multiplying all the "vectors"

	        c[px][i] * |px> + c[py][i] * |py> + c[pz][i] * |pz>

       by the matrix R

               / cos(phi) cos(theta)   sin(phi) cos(theta)  -sin(theta) \
          R = |     -sin(phi)                cos(phi)            0       |
               \ cos(phi) sin(theta)   sin(phi) sin(theta)   cos(theta) /

       whereby we get the LCAO coefficients in the local basis

            c[px'][i] * |px'> + c[py'][i] * |py'> + c[pz'][i] * |pz'>

       First we've got to calculate the sines und cosines of the angles
       phi and theta ... */

	ct = ( M->z - L->z ) / RLM;
	st = sqrt( 1.0 - sqr( ct ) );

	if ( fabs( ct ) > 0.9999999 )      /* if both atoms lay along the z-axis */
	{
		cp = 1.0;
		sp = 0.0;
	}
	else
	{
		temp = 1.0 / sqrt( sqr( M->x - L->x ) + sqr( M->y - L->y ) );

		cp = ( M->x - L->x ) * temp;
		sp = ( M->y - L->y ) * temp;
	}

	/* ...then set up the rotation matrix ... */

	R[ X ][ X ] =   cp * ct;
	R[ X ][ Y ] =   sp * ct;
	R[ X ][ Z ] = - st;
	R[ Y ][ X ] = - sp;
	R[ Y ][ Y ] =   cp;
	R[ Y ][ Z ] =   0.0;
	R[ Z ][ X ] =   cp * st;
	R[ Z ][ Y ] =   sp * st;
	R[ Z ][ Z ] =   ct;

	/* ... and calculate the the LCAO coefficients for atom M in the
	   local system - but take care : because of the use of elliptical
	   coordinates for the calculation of the integrals a left hand
	   system is needed on atom M system - the z-axis on atom L points
	   to atom M, while the z-axis on atom M points to atom L, x- and
	   y-axis being identical on both atoms ! For this reason the z
	   LCAO coefficient is simply multiplied by -1, reflecting the
	   inversion of the z-axis for the system at atom M. */

	clM[  S ]  =  M->c[  S ][ i ];
	clM[ PX ] =   M->c[ PX ][ i ] * R[ X ][ X ]
	 	        + M->c[ PY ][ i ] * R[ X ][ Y ]
			    + M->c[ PZ ][ i ] * R[ X ][ Z ];
	clM[ PY ] =   M->c[ PX ][ i ] * R[ Y ][ X ]
                + M->c[ PY ][ i ] * R[ Y ][ Y ];
	clM[ PZ ] = - M->c[ PX ][ i ] * R[ Z ][ X ]
                - M->c[ PY ][ i ] * R[ Z ][ Y ]
                - M->c[ PZ ][ i ] * R[ Z ][ Z ];
}


/*************************************************/
/* Calculates the auxilary intergrals A and B.   */
/*                                               */
/*  ->                                           */
/*     1. interatomic distance                   */
/*     2. Slater orbital exponent of first atom  */
/*     3. Slater orbital exponent of second atom */
/*     4. array of A intergrals                  */
/*     5. array of B integrals                   */
/*************************************************/

static void make_aux_ints( double R, double zeta1, double zeta2,
						   double *AI, double *BI )
{
	make_A_ints( 0.5 * R * ( zeta1 + zeta2 ), AI );
	make_B_ints( 0.5 * R * ( zeta1 - zeta2 ), BI );
}


/***************************************************************************/
/* Here the set of A integrals needed in the calculation of the two-center */
/* integrals is computed. No special care has taken for the case p = 0.    */
/*                                                                         */
/* Input parameter : Exponential parameter p                               */
/*                   Pointer to the values of the integrals                */
/***************************************************************************/

static void make_A_ints(double p, double *AI)
{
    AI[ 0 ] = exp( -p ) / p;
    AI[ 1 ] = AI[ 0 ] / p + AI[ 0 ];
    AI[ 2 ] = 2.0 * AI[ 1 ] / p + AI[ 0 ];
    AI[ 3 ] = 3.0 * AI[ 2 ] / p + AI[ 0 ];
    AI[ 4 ] = 4.0 * AI[ 3 ] / p + AI[ 0 ];
}


/***************************************************************************/
/* Here the set of B integrals needed in the calculation of the two-center */
/* integrals is computed. Special care has to be taken for the case q = 0. */
/*                                                                         */
/* Input parameter : Exponential parameter q                               */
/*                   Pointer to the values of the integrals                */
/***************************************************************************/

static void make_B_ints( double q, double *BI )
{
    double csh;


    if ( fabs( q ) < 5.0e-3 )                 /* take care for small q's ! */
    {
        BI[ 0 ] =  2.0;
        BI[ 1 ] = -2.0 * q / 3.0;
        BI[ 2 ] =  2.0 / 3.0;
        BI[ 3 ] = -0.4 * q;
        BI[ 4 ] =  0.4;
    }
    else
    {
        csh = 2.0 * cosh( q );

        BI[ 0 ] = 2.0 * sinh( q ) / q;
        BI[ 1 ] = ( BI[ 0 ] - csh ) / q;
        BI[ 2 ] = 2.0 * BI[ 1 ] / q + BI[ 0 ];
        BI[ 3 ] = ( 3.0 * BI[ 2 ] - csh) / q;
        BI[ 4 ] = 4.0 * BI[ 3 ] / q + BI[ 0 ];
    }
}
