/***********************************************************************

  $Id: paramag.c,v 1.6 2000/11/04 13:45:40 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  This file contains routines for the calculation of the paramagnetic
  part of the g-tensor - see also the file "param_u.c".

***********************************************************************/



#include "g_rhf.h"                  /* definitions and declarations */



/* Declaration of routines local to this file */


static double lev( int coord, DMATRIX s, int i, int j );
static double mc_lev_s( int l, int K, ATOM *atom, int atoms, double *R1,
						DMATRIX R2, double *R3, DMATRIX R4 );
static double vprod( int a, double *r, DMATRIX R, int L );



/****************************************************************************/
/* Calculates paramagmetic contributions to the G tensor for RHF functions. */
/*                                                                          */
/* ->                                                                       */
/*     1. matrix of the paramagnetic contribution to G                      */
/*     2. array of structures for all atoms                                 */
/*     3. array of orbital energies                                         */
/*     4. pointer to structure with number of atoms, states etc.            */
/*     5. pointer to structure with command line options etc.               */
/****************************************************************************/

int calculate_paramagnetic_part( DMATRIX g3, ATOM *atom, double *energy,
								 NUMS *n, OPTION *options )
{
	double sum_L, sum_K, ee, temp, R1[ 3 ], R3[ 3 ];
	int i, j, l, m, K;
	double *JI = NULL,
		   *KI;
	DMATRIX R2 = NULL, R4 = NULL;
	char *command;


	/* First, if necessary, the energy corrections are calculated */

	if ( options->dec &&
		! energy_corrections( &JI, &KI, atom, energy, n, options ) )
		return 0;

	if ( ! options->dec && options->le )
		log_energy( options->outfile, energy, n, options );

	/* The log file is opened */

	options->fp = open_log_file( options );

	/* Now really start to calculate the paramagnetic contribution : begin
	   with contributions from excitations starting from the SOMO */

	for ( j = n->occ_alpha; j < n->states; j++ )
	{
		/* Start by doing most of the work needed for the calculation of the
		   two- and three-center contributions to the orbital momenta */

		if ( options->mc && ! prep_mc( R1, &R2, R3, &R4, atom, n->atoms,
									   n->SOMO, j, options ) )
		{
			if ( options->dec )
				free_energy_corr_memory( JI );
			return 0;
		}

		for ( l = X; l <= Z; l++ )        /* loop over all x, y and z */
			for ( m = X; m <= Z; m++ )    /* again loop over x, y and z */
			{
				/* The nominator is calculated with the method depending on
				   either the inclusion of two- and three-center terms or use
				   of Stone's approximation */

				if ( options->mc )
					for ( sum_K = 0.0, K = 0; K < n->atoms; K++ )
						sum_K += 
							mc_lev_s( l, K, atom, n->atoms, R1, R2, R3, R4 )
							* atom[ K ].xi * lev( m, atom[ K ].c, n->SOMO, j );
				else
				{
					for ( sum_K = sum_L = 0.0, K = 0; K < n->atoms; K++ )
					{
						sum_L += lev( l, atom[ K ].c, n->SOMO, j );
						sum_K += atom[ K ].xi
							     * lev( m, atom[ K ].c, n->SOMO, j );
					}
					sum_K *= sum_L;
				}

				/* The excitation energy is calculated... */

				ee = energy[ j ] - energy[ n->SOMO ];
				if ( options->dec )
					ee += - JI[ j ] + 0.5 * KI[ j ] + 0.5 * JI[ n->SOMO ];

				/* ...and finally the contribution to G from the excitation
                   from the SOMO into the unoccupied state j */

				temp = ( options->Stone || options->half ? 0.5 : 1.0 )
					   * sum_K / ee;
			    g3[ l ][ m ] -= temp;

				if ( options->fp )
					fprintf( options->fp, "%c%c: %3d->%3d % 8.4f -> % 9.6f\n",
							 l + 'x', m + 'x', n->occ_alpha, j + 1,
							 au_to_eV * ee, - temp * g_e * sqr( alpha ) );
			}
	}

	/* Now calculate contributions of excitations into the SOMO */

	for ( i = 0; i < n->occ_beta; i++ )
	{
		/* Again start by preparing the calculation of the  two- and
		   three-center contributions to the orbital momenta */

		if ( options->mc && ! prep_mc( R1, &R2, R3, &R4, atom, n->atoms, i,
									   n->SOMO, options ) )
        {
			if ( options->dec )
				free_energy_corr_memory( JI );
			return 0;
		}

		for ( l = X; l <= Z; l++ )        /* loop over all x, y and z */
			for ( m = X; m <= Z; m++ )    /* again loop over x, y and z */
			{
				/* The nominator is calculated with the method depending on
				   either the inclusion of two- and three-center terms or use
				   of Stone's approximation */

				if ( options->mc )
					for ( sum_K = 0.0, K = 0; K < n->atoms; K++ )
						sum_K +=
							mc_lev_s( l, K, atom, n->atoms, R1, R2, R3, R4 )
							* atom[ K ].xi * lev( m, atom[ K ].c, i, n->SOMO );
				else
				{
					for ( sum_K = sum_L = 0.0, K = 0; K < n->atoms; K++ )
					{
						sum_L += lev( l, atom[ K ].c, i, n->SOMO );
						sum_K += atom[ K ].xi
							     * lev( m, atom[ K ].c, i, n->SOMO );
					}
					sum_K *= sum_L;
				}

				/* The excitation energy is calculated... */

				ee = energy[ n->SOMO ] - energy[ i ];
				if ( options->dec )
					ee += - JI[ i ] + 0.5 * KI[ i ] + 0.5 * JI[ n->SOMO ];

				/* ...and finally the contribution to G from the excitations 
                   from the doubly occupied orbital i into the SOMO */

				temp = ( options->Stone || options->half ? 0.5 : 1.0 )
					   * sum_K / ee;
				g3[ l ][ m ] += temp;

				if ( options->fp )
					fprintf( options->fp, "%c%c: %3d->%3d % 8.4f -> % 9.6f\n",
							 l + 'x', m + 'x', i + 1, n->occ_alpha,
							 au_to_eV * ee, temp * g_e * sqr( alpha ) );
			}
	}

	/* Close log file and sort it */

	if ( options->fp )
	{
		fclose( options->fp );
		command = malloc( strlen( options->param_dir ) +
						  strlen( options->outfile ) + 18 );
		if ( command != NULL )
		{
			strcpy( command, options->param_dir );
			strcat( command, "g_lpsort " );
			strcat( command, options->outfile );
			strcat( command, ".paramag" );
			options->fp = popen( command, "w" );
			pclose( options->fp );
			free( command );
		}
	}

	/* Finally memory used for energy corrections etc. is freed */

	free_energy_corr_memory( JI );
	if ( options->mc )
	{
		free( R2 );
		free( R4 );
	}

	return 1;
}


/*****************************************************************/
/* Calculates component of expectation value of orbital momentum */
/* using only one-center terms                                   */
/* ->                                                            */
/*    1. cartesian component of the orbital momentum             */
/*    2. LCAO coefficients of the actual atom                    */
/*    3. state on left side of braket                            */
/*    4. state on right side of braket                           */
/* <- expectation value of orbital momentum                      */
/*****************************************************************/

static double lev( int coord, DMATRIX s, int i, int j )
{
	int c1, c2;

	c1 = ( ( coord + 2 ) % 3 ) + PX;
	c2 = ( ( coord + 1 ) % 3 ) + PX;

	return s[ c1 ][ i ] * s[ c2 ][ j ] - s[ c2 ][ i ] * s[ c1 ][ j ];
}


/***************************************************************************/
/* Calculates the expectation value of the orbital momentum including two- */
/* and three-center terms. Most of the work connected with the calculation */
/* is already done in the routine prep_mc(), which must be called earlier. */
/* ->                                                                      */
/*    1. cartesian component of the orbital momentum                       */
/*    2. number of the atom the orbital momentum refers to                 */
/*    3. array of structures for all atoms                                 */
/*    4. number of atoms                                                   */
/*    5. arrays and matrices with the pre-processed values                 */
/* <- expectation value of orbital momentum                      */
/***************************************************************************/

static double mc_lev_s( int l, int K, ATOM *atom, int atoms, double *R1,
						DMATRIX R2, double *R3, DMATRIX R4 )
{
	double sum, rKL[ 3 ];
	int L;

	sum = R1[ l ] + R3[ l ];
	
	for ( L = 0; L < atoms; L++ )
	{
		if ( K == L )
			continue;

		rKL[ X ] = atom[ L ].x - atom[ K ].x;
		rKL[ Y ] = atom[ L ].y - atom[ K ].y;
		rKL[ Z ] = atom[ L ].z - atom[ K ].z;

		sum += vprod( l, rKL, R2, L );
		if ( R4 != NULL )
			sum += vprod( l, rKL, R4, L );
	}

	return sum;
}


/***************************************************************************/
/* Calculates one of the components of the vector products from mc_lev_s() */
/***************************************************************************/

static double vprod( int a, double *r, DMATRIX R, int L )
{
	int b, c;


    b = ( a + 1 ) % 3;
    c = ( a + 2 ) % 3;

	return r[ b ] * R[ c ][ L ] - r[ c ] * R[ b ][ L ];
}
