/********************************************************************

 $Id: read.c,v 1.9 2000/11/25 14:38:55 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

 This file contains routines to read in the data from the input file

*********************************************************************/


#include "g_rhf.h"          /* definitions and declarations */



/* Declaration of routines local to this file */

static int get_mem( NUMS n, ATOM **atom, double **energy );
static int get_numbers( NUMS *n, OPTION *options );
static int get_coeff( ATOM *atom, double *energy, NUMS n, OPTION *options );
static int get_coordinates( int atoms, ATOM *atom, OPTION *options,
							PARAMS *params );
static int set_atom_params( ATOM *atom, char *name, PARAMS *params );
static void normalize( ATOM *atom, NUMS n );


/*************************************************************************/
/* Reads all data needed by the program from the data file, allocates    */
/* needed and calculates several often used determinants.                */
/*  ->                                                                   */
/*     1. pointer to structure for number of atoms etc.                  */
/*     2. pointer arrray of structures for atoms                         */
/*     3. pointer to array of orbital energies                           */
/*     4. array of structures for atomic parameter                       */
/*     5. structure with options                                         */
/*  <- 1 : everything ok, 0 : error detected                             */
/*************************************************************************/

int get_all_data( NUMS *n, ATOM **atom, double **energy, PARAMS *params,
				  OPTION *options )
{
    /* Read number of electrons and atoms etc. */

    if ( ! get_numbers( n, options ) )
        return 0;


    /* Test if this can be a doublet-state */

    if ( n->occ_alpha - n->occ_beta != 1 )
    {
        printf( "This is not a doublet groundstate!\n" );
        return 0;
    }

    /* Allocate memory for atom structures and energies */

    if ( ! get_mem( *n, atom, energy ) )
        return 0;

    /* Read in the coordinates */

    if ( ! get_coordinates( n->atoms, *atom, options, params ) )
    {
        clear( *energy, *atom, n->atoms );
        return 0;
    }

    /* Read LCAO-coefficients and energies */

    if ( ! get_coeff( *atom, *energy, *n, options ) )
    {
        clear( *energy, *atom, n->atoms );
        return 0;
    }

	return 1;
}


/***********************************************************************/
/* Allocates memory for orbital energies, structures for the atoms and */
/* the matrices of LCAO-coefficients of the atomic orbitals.           */
/*                                                                     */
/* Input parameter : Structure with number of states, atoms etc.       */
/*                   Pointer to pointer to structures for atoms        */
/*                   Pointer to pointer to orbital energies            */
/*                                                                     */
/* Return value    : 1 everything ok, 0 not enough memory              */
/***********************************************************************/

static int get_mem( NUMS n, ATOM **atom, double **energy )
{
    int i, j;                   /* counter */


    /* Allocate memory for the orbital energies */

    if ( ( *energy = calloc( n.states, sizeof( double ) ) ) == NULL )
    {
        printf( "Not enough memory for energies!\n" );
        return 0;
    }

    /* Also allocate memory for a structure for every atom */

    if ( ( *atom = calloc( n.atoms, sizeof( ATOM ) ) ) == NULL )
    {
        printf( "Not enough memory for atoms!\n" );
        free( *energy );
        return 0;
    }

    /* Allocate memory for the LCAO coefficients for every atom */

    for ( i = 0; i < n.atoms; i++ )
    {
        ( *atom )[ i ].c = dmatrix( 4, n.states );

        if ( ( *atom)[ i ].c == NULL )
        {
            printf( "Out of memory at %d. of %d atoms !\n", i + 1,
					n.atoms + 1 );
            for ( j = 0; j <= i ; j++ )
                if ( ( *atom )[ j ].c )
                    free( ( *atom )[ j ].c );

            free( *atom );
            free( *energy );
            return 0;
        }
    }

    return 1;
}


/*********************************************************************/
/* Reads number of states, atoms, electrons and occupied states from */
/* the data file.                                                    */
/*                                                                   */
/* Input parameter : Pointer to structure for number of atoms etc.   */
/*                   Structure with options                          */
/*                                                                   */
/* Return value    : 1 everything ok, 0 error detected               */
/*********************************************************************/

static int get_numbers( NUMS *n, OPTION *options )
{
	char *command;
	FILE *fp;


	command = malloc( strlen( options->param_dir ) +
					  strlen( options->infile ) + 20 );

	if ( command == NULL )
	{
		printf( "Running out of memory !\n" );
		return 0;
	}

	sprintf( command, "%snumbers.pl %s", options->param_dir, options->infile );

	if ( ( fp = popen( command, "r" ) ) == NULL )
	{
		free( command );
		printf( "Can't read data file !\n" );
		return 0;
	}

	free( command );

	if ( fscanf( fp, "%d %d %d %d %d", &n->atoms, &n->elecs, &n->states,
				 &n->occ_beta, &n->occ_alpha ) != 5 )
	{
		printf( "Can't read data file !\n" );
		pclose( fp );
		return 0;
	}

	pclose( fp );

	n->SOMO = n->occ_alpha - 1;

	return 1;
}


/***********************************************************************/
/* Reads in energies and LCAO coefficients from input data file. The   */
/* function get_coordinates() has to be called before this function is */
/* called !                                                            */
/*                                                                     */
/* Input parameter : Pointer to structures for all atoms               */
/*                   Pointer to array of orbital energies              */
/*                   strtucture for number of states, atoms etc.       */
/*                   Pointer to structures of parameters               */
/*                   Structure with options                            */
/*                                                                     */
/* Return value    : 1 everything ok, 0 error detected                 */
/***********************************************************************/

static int get_coeff( ATOM *atom, double *energy, NUMS n, OPTION *options )
{
	FILE *fp;
	char *command;
	int i, j, K, type, fields, num, set;


	/* Get all orbital energies */

	command = malloc( ( strlen( options->param_dir ) +
						strlen( options->infile ) + 21 ) * sizeof( char ) );
	if ( command == NULL )
	{
		printf( "Running out of memory !\n" );
		return 0;
	}

	sprintf( command, "%senergies.pl %s",
			 options->param_dir, options->infile );

	if ( ( fp = popen( command, "r" ) ) == NULL )
	{
		free( command );
		printf( "Can't read data file!\n" );
		return 0;
	}

	for ( i = 0; i < n.states; i++ )
	{	
		if ( fscanf( fp, "%lf", energy + i ) != 1 )
		{
			free( command );
			printf( "Can't read orbital energies !\n" );
			pclose( fp );
			return 0;
		}
		energy[ i ] /= au_to_eV;
	}

	pclose( fp );


	/* Get all LCAO coefficients */

	sprintf( command, "%slcaos.pl %s", options->param_dir, options->infile );

	if ( ( fp = popen( command, "r" ) ) == NULL )
	{
		free( command );
		printf( "Can't read data file!\n" );
		return 0;
	}

	free( command );

	while ( fscanf(fp, "%d %d %d %d", &set, &fields, &type, &num ) == 4 )
	{
		/* We first got to find the position in the array of atom structures
		   of the atom with number 'num' because in some input files the atom
		   numbers are not continous */

		for ( j = 0; j < n.atoms; j++ )
			if ( atom[ j ].atom_no == num )
				break;
		if ( j >= n.atoms )
		{
			printf( "Found LCAO for unknown atom #%d\n", num );
			return 0;
		}

		for ( i = set - fields; i < set; i++ )
			fscanf( fp, "%lf", &atom[ j ].c[ type ][ i ] );
	}

	pclose( fp );
	
	/* Make sure the LCAO coeffizients for the p orbitals are zero for
	   atoms with only a 1s orbital (i.e. hydrogen) */

	for ( K = 0; K < n.atoms; K++ )
		if ( atom[ K ].n == 1 )
			for ( i = 0; i < n.states; i++ )
				atom[ K ].c[ PX ][ i ] = atom[ K ].c[ PY ][ i ]
					                   = atom[ K ].c[ PZ ][ i ] = 0.0;

	/* Finally renormalize LCAO's */

	normalize( atom, n );

    /* Return succesfully */

    return 1;
}


/****************************************************************************/
/* Reads in the cartesian atomic coordinates - if there are two tables with */
/* cartesian coordinates (e. g. in files where a geometry optimization was  */
/* done) the second table is used.                                          */
/*                                                                          */
/* Input parameter : Number of atoms                                        */
/*                   Pointer to structures for all atoms                    */
/*                   Structure with options                                 */
/*                   Array of structures for atomic parameter               */
/*                                                                          */
/* Return value    : 1 everything ok, 0 error detected                      */
/****************************************************************************/

static int get_coordinates( int atoms, ATOM *atom, OPTION *options,
							PARAMS *params )
{
	FILE *fp;
	char *command, name[10];
	int i, num;
	double x, y, z;


	command = malloc( strlen( options->param_dir ) +
					  strlen( options->infile ) + 20 );
	if ( command == NULL )
	{
		printf( "Running out of memory !\n" );
		return 0;
	}

	sprintf( command, "%scoordina.pl %s",
			 options->param_dir, options->infile );

	if ( ( fp = popen( command, "r" ) ) == NULL )
	{
		free( command );
		printf( "Can't read data file!\n" );
		return 0;
	}

	free( command );

	for ( i = 0; i < atoms; i++ )
	{
		if ( fscanf( fp, "%d %s %lf %lf %lf", &num, name, &x, &y, &z ) != 5 )
		{
			printf( "Can't read coordinates at i = %d!\n", i );
			pclose( fp );
			return 0;
		}

		atom[ i ].atom_no = num;
		if ( ! set_atom_params( &atom[ i ], name, params ) )
		{
			pclose( fp );
			return 0;
		}

		atom[ i ].x = x / R_BOHR;
		atom[ i ].y = y / R_BOHR;
		atom[ i ].z = z / R_BOHR;
	}

	pclose( fp );

	return 1;
}


/**************************************************************/
/* Sets the atomic parameters for the actual parameterization */ 
/* Input parameter : Pointer to structures for actual atom    */
/*                   Type of atom (i.e. string with name)     */
/*                   Array of structures for atomic parameter */
/*                                                            */
/* Return value    : 1 everything ok, 0 error detected        */
/**************************************************************/

static int set_atom_params( ATOM *atom, char *name, PARAMS *params )
{
	int atom_type = -1;


	if ( ! strcmp( name, "H" ) )
	{
		atom_type     = Hydrogen;
		atom->n       = 1;
		atom->sigma_s = Z_H * params[ Hydrogen ].zeta_s;
		atom->sigma_p = 0.0;
		atom->xi      = params[ Hydrogen ].xi;
	}

	if ( ! strcmp( name, "C" ) )
	{
		atom_type     = Carbon;
		atom->n       = 2;
		atom->sigma_s = 0.5 * Z_C * params[ Carbon ].zeta_s;
		atom->sigma_p = 0.5 * Z_C * params[ Carbon ].zeta_p;
		atom->xi      = params[ Carbon ].xi;
	}

	if ( ! strcmp( name, "N" ) )
	{
		atom_type     = Nitrogen;
		atom->n       = 2;
		atom->sigma_s = 0.5 * Z_N * params[ Nitrogen ].zeta_s;
		atom->sigma_p = 0.5 * Z_N * params[ Nitrogen ].zeta_p;
		atom->xi      = params[ Nitrogen ].xi;
	}

	if ( ! strcmp( name, "O" ) )
	{
		atom_type     = Oxygen;
		atom->n       = 2;
		atom->sigma_s = 0.5 * Z_O * params[ Oxygen ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Oxygen ].zeta_p;
		atom->xi      = params[ Oxygen ].xi;
	}

	if ( ! strcmp( name, "F" ) )
	{
		atom_type     = Fluorine;
		atom->n       = 2;
		atom->sigma_s = 0.5 * Z_O * params[ Fluorine ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Fluorine ].zeta_p;
		atom->xi      = params[ Fluorine ].xi;
	}

	if ( ! strcmp( name, "Na" ) || ! strcmp( name, "NA" ) )
	{
		atom_type     = Sodium;
		atom->n       = 3;
		atom->sigma_s = 0.5 * Z_O * params[ Sodium ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Sodium ].zeta_p;
		atom->xi      = params[ Sodium ].xi;
	}

	if ( ! strcmp( name, "Mg" ) || ! strcmp( name, "MG" ) )
	{
		atom_type     = Magnesium;
		atom->n       = 3;
		atom->sigma_s = Z_Mg * params[ Magnesium ].zeta_s / 3.0;
		atom->sigma_p = Z_Mg * params[ Magnesium ].zeta_p / 3.0;
		atom->xi      = params[ Magnesium ].xi;
	}

	if ( ! strcmp( name, "P" ) )
	{
		atom_type     = Phosphorus;
		atom->n       = 3;
		atom->sigma_s = 0.5 * Z_O * params[ Phosphorus ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Phosphorus ].zeta_p;
		atom->xi      = params[ Phosphorus ].xi;
	}

	if ( ! strcmp( name, "S" ) )
	{
		atom_type     = Sulfur;
		atom->n       = 3;
		atom->sigma_s = 0.5 * Z_O * params[ Sulfur ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Sulfur ].zeta_p;
		atom->xi      = params[ Sulfur ].xi;
	}

	if ( ! strcmp( name, "Cl" ) || ! strcmp( name, "CL" ) )
	{
		atom_type     = Chlorine;
		atom->n       = 3;
		atom->sigma_s = 0.5 * Z_O * params[ Chlorine ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Chlorine ].zeta_p;
		atom->xi      = params[ Chlorine ].xi;
	}

	if ( ! strcmp( name, "Zn" ) || ! strcmp( name, "ZN" ) )
	{
		atom_type     = Zinc;
		atom->n       = 4;
		atom->sigma_s = 0.5 * Z_O * params[ Zinc ].zeta_s;
		atom->sigma_p = 0.5 * Z_O * params[ Zinc ].zeta_p;
		atom->xi      = params[ Zinc ].xi;
	}

	if ( atom_type < Hydrogen )
	{
		printf( "Unknown atom type (%s) of atom in data file!\n", name );
		return 0;
	}

	if ( atom->xi < 0 )
	{
		printf( "Spin-orbit coupling for %s unknown, using 0.\n", name );
		atom->xi = 0.0;
	}

	atom->Gss    = params[ atom_type ].Gss;
	atom->Gsp    = params[ atom_type ].Gsp;
	atom->Gpp    = params[ atom_type ].Gpp;
	atom->Gp2    = params[ atom_type ].Gp2;
	atom->Hsp    = params[ atom_type ].Hsp;
	atom->zeta_s = params[ atom_type ].zeta_s;
	atom->zeta_p = params[ atom_type ].zeta_p;
	atom->D1     = params[ atom_type ].D1;
	atom->D2     = params[ atom_type ].D2;
	atom->rho_0  = params[ atom_type ].rho_0;
	atom->rho_1  = params[ atom_type ].rho_1;
	atom->rho_2  = params[ atom_type ].rho_2;

	if ( atom->n > 1 )
		atom->tc_fac = 8.0 * sqrt( 3.0 )
			* ( atom->zeta_s - atom->zeta_p / 3.0 )
				* pow( sqrt( atom->zeta_s * atom->zeta_p )
					  / ( atom->zeta_s + atom->zeta_p ), 5.0 );

	return 1;
}


/********************************************************************/
/* Normalizes the eigenvectors to avoid unnecessary rounding errors */
/*  ->                                                              */
/*     1. array of structures for all atoms                         */
/*     2. structure for number of atoms, states etc.                */
/********************************************************************/

static void normalize( ATOM *atom, NUMS n )
{
	int i, l, K;
	double p;


	for ( i = 0; i < n.states; i++ )
	{
		for ( p = 0.0, K = 0; K < n.atoms; K++ )
		    for ( l = S; l <= PZ; l++ )
				p += sqr( atom[ K ].c[ l ][ i ] );

		p = 1.0 / sqrt( p );

		for ( K = 0; K < n.atoms; K++ )
			for ( l = S; l <= PZ; l++ )
				atom[ K ].c[ l ][ i ] *= p;
	}
}
