/**************************************************************************

  $Id: m_alloc.c,v 1.3 2001/08/19 20:56:10 jens Exp $

  Copyright (C) 2001 Jens Thoms Toerring

  g_rhf is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  g_rhf is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with g_rhf; see the file COPYING.  If not, write to
  the Free Software Foundation, 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.

  In the book "Numerical Recipes in C" (2nd ed.) by W. H. Press, S. A.
  Teukolsky, W. T. Vetterling and B. P. Flannery (Cambridge University
  Press 1992) a method is proposed to overcome some disadvantages of
  the C programming language concerning the handling of matrices.
  As the authors point out C poses a problem with passing matrices of
  variable size to subroutines. I.e., a construction like

  main()
  {
  	  double matrix[n][m];
  	  ...
  	  function(a, n, m);
  	  ...
  }

  function(double a[][m], int n, int m)
  {
  	  ...
  	  a[i][j] = ...
  	  ...
  }

  is unfortunatly not a legal construction - C requires that the num-
  ber of columns of the matrix is a constant known at compilation time.
  A possible way to handle this problem is to treat the matrix in the
  subroutine as a linear array and do the bookkeeping of indices 'by
  hand', e. g.

  function(double *a, int n, int m)
  {
  	  ...
  	  *(a + i * m + j) = ...
  	  ...
  }

  Though this is a legal way to overcome the problem it has its draw-
  backs. Beside being rather liable to errors and hard to read, it is
  a rather unconvenient method for people not being system programers
  or pointer enthusiasts and prefer the matrix notation over this
  'cryptic' way to access the elements of a matrix.

  The method proposed by the authors of "Numerical Recipes" exploits the
  different ways C treats two-dimensional arrays and pointers to arrays
  of pointers. The expression a[i][j] will be translated differently by
  the compiler in these two cases :

  		double a[n][m];  ->  a[i][j]	translates to  *(a + i * m + j)
  while
  		double **a;	   ->  a[i][j]	translates to  *(*(a + i) + j)

  This difference can be used to implement a solution to the above men-
  tioned problem at the expense of some additional storage required :
  Instead of allocating memory just for the matrix alone, storage is
  allocated for the matrix as well as for an array of pointers to the
  rows of the matrix (i. e., the first elements of the rows). The only
  additional work to be done is to initialize this array of pointers.
  Afterwards the matrix can be easily passed to subroutines as pointer
  to a pointer and the matrix notation can be used also in the subrou-
  tine. There is even a slight increase in computational speed since to
  compute the memory location of the element a[i][j] of a two-dimen-
  sional array takes two additions plus a multiplication while with the
  method presented here only two additions are needed.

  It is also possible to do the 'classic' pointer arithmetics when one
  keeps in mind that *a = a[0] instead of a itself is the pointer to
  the first element of the matrix - the following elements are orderd
  in the usual way, i. e., the matrix is stored by rows.

  The subroutines in this file do all the necessary work to allocate
  memory for a matrix and the array of pointers to its rows as well as
  to initialize this array. Just call the function for the needed type
  of matrix with the number of rows and columns as the parameters. A
  pointer to a pointer to the first element of the matrix will be re-
  turned (or NULL, if there is not enough memory left or the arguments
  are incorrect, i. e. less than one).

  	   To deallocate the memory used for the matrix no special
  		   function is needed, the usual free() will do !

  In addition to this file there is a header file m_alloc.h which con-
  tains the function declarations to be included into modules using
  this functions.

  Contents :

  		DMATRIX dmatrix()		  Make a matrix of type double
  		FMATRIX fmatrix()		  Make a matrix of type float
  		LMATRIX lmatrix()		  Make a matrix of type long
  		IMATRIX imatrix()		  Make a matrix of type int

*************************************************************************/


#include "m_alloc.h"


DMATRIX dmatrix( long rows, long columns )
{
	DMATRIX m;
	long i;


	/* Check for pathological cases (there is no matrix with less than
	   one element) */

	if ( rows < 1 || columns < 1 )
		return NULL;

	/* Allocate memory for the matrix as well as for an array of pointers
	   to its rows - return NULL pointer on failure */

	m = malloc( ( size_t ) rows * ( ( size_t ) columns *
									sizeof( double ) + sizeof( double * ) ) );

	if ( m == NULL )
		return NULL;								 /* Allocation failed */

	/* Initialize the array of pointers to the rows of the matrix */

	m[ 0 ] = ( double * ) ( m + rows );
	for ( i = 1; i < rows; i++ )
		m[ i ] = m[ i - 1 ] + columns;

	/* Return pointer to the pointer array to the rows of the matrix */

	return m;
}


FMATRIX fmatrix( long rows, long columns )
{
	FMATRIX m;
	long i;


	/* Check for pathological cases (there is no matrix with less than
	   one element) */

	if ( rows < 1 || columns < 1 )
		return NULL;

	/* Allocate memory for the matrix as well as for an array of pointers
	   to its rows - return NULL pointer on failure */

	m = malloc( ( size_t ) rows * ( ( size_t ) columns * sizeof( float )
									+ sizeof( float * ) ) );

	if ( m == NULL )
		return NULL;								 /* Allocation failed */

	/* Initialize the array of pointers to the rows of the matrix */

	m[ 0 ] = ( float * ) ( m + rows );
	for ( i = 1; i < rows; i++ )
		m[ i ] = m[ i - 1 ] + columns;

	/* Return pointer to the pointer array to the rows of the matrix */

	return m;
}


LMATRIX lmatrix( long rows, long columns )
{
	LMATRIX m;
	long i;


	/* Check for pathological cases (there is no matrix with less than
	   one element) */

	if ( rows < 1 || columns < 1 )
		return NULL;

	/* Allocate memory for the matrix as well as for an array of pointers
	   to its rows - return NULL pointer on failure */

	m = malloc( ( size_t ) rows * ( ( size_t ) columns * sizeof( long )
									+ sizeof( long * ) ) );

	if ( m == NULL )
		return NULL;								 /* Allocation failed */

	/* Initialize the array of pointers to the rows of the matrix */

	m[ 0 ] = ( long * ) ( m + rows );
	for ( i = 1; i < rows; i++ )
		m[ i ] = m[ i - 1 ] + columns;

	/* Return pointer to the pointer array to the rows of the matrix */

	return m;
}


IMATRIX imatrix( long rows, long columns )
{
	IMATRIX m;
	long i;


	/* Check for pathological cases (there is no matrix with less than
	   one element) */

	if ( rows < 1 || columns < 1 )
		return NULL;

	/* Allocate memory for the matrix as well as for an array of pointers
	   to its rows - return NULL pointer on failure */

	m = malloc( ( size_t ) rows * ( ( size_t ) columns * sizeof( int )
									+ sizeof( int * ) ) );

	if ( m == NULL )
		return NULL;								 /* Allocation failed */

	/* Initialize the array of pointers to the rows of the matrix */

	m[ 0 ] = ( int * ) ( m + rows );
	for ( i = 1; i < rows; i++ )
		m[ i ] = m[ i - 1 ] + columns;

	/* Return pointer to the pointer array to the rows of the matrix */

	return m;
}


CMATRIX cmatrix( long rows, long columns )
{
	CMATRIX m;
	long i;


	/* Check for pathological cases (there is no matrix with less than
	   one element) */

	if ( rows < 1 || columns < 1 )
		return NULL;

	/* Allocate memory for the matrix as well as for an array of pointers
	   to its rows - return NULL pointer on failure */

	m = malloc( ( size_t ) rows * ( ( size_t ) columns + sizeof( char * ) ) );

	if ( m == NULL )
		return NULL;								 /* Allocation failed */

	/* Initialize the array of pointers to the rows of the matrix */

	m[ 0 ] = ( char * ) ( m + rows );
	for ( i = 1; i < rows; i++ )
		m[ i ] = m[ i - 1 ] + columns;

	/* Return pointer to the pointer array to the rows of the matrix */

	return m;
}
